#!/usr/bin/env python3
#
# Created using scripts/create-bip-hd-chain-params.py
# Source data:
#   https://github.com/MetaMask/slip44/blob/main/slip44.json (1bc984bee)
#   https://github.com/ebellocchia/bip_utils (5649541c6)

from collections import namedtuple

def parse_data():

	_d = namedtuple(
		'bip_hd_data',
		'idx chain curve network addr_cls vb_prv vb_pub vb_wif vb_addr def_path name')
	_u = namedtuple(
		'bip_hd_data_partial',
		'idx chain name')

	def parse_line(line):
		l = line.split()

		if l[2] == '-':
			return _u(
				idx   = int(l[0]),
				chain = l[1],
				name  = ' '.join(l[3:]),
			)
		else:
			return _d(
				idx      = int(l[0]),
				chain    = l[1],
				curve    = defaults.curve if l[2] == 'x' else l[2],
				network  = 'mainnet' if l[3] == 'm' else 'testnet' if l[3] == 'T' else None,
				addr_cls = l[4],
				vb_prv   = defaults.vb_prv if l[5] == 'x' else l[5],
				vb_pub   = defaults.vb_pub if l[6] == 'x' else l[6],
				vb_wif   = l[7],
				vb_addr  = l[8],
				def_path = defaults.def_path if l[9] == 'x' else l[9],
				name     = ' '.join(l[10:]),
			)

	out = {}
	for line in _data_in.strip().splitlines():
		if not line or line.startswith('IDX'):
			continue
		if line.startswith('['):
			key = line[1:-1]
			continue
		p = parse_line(line)
		if key in out:
			out[key][p[1]] = p
		elif key == 'defaults':
			out['defaults'] = p
			defaults = p
		else:
			out[key] = {p[1]: p}

	return out

# RUNE derivation is SLIP-10, not BIP-44, but we treat them as equivalent

_data_in = """

[defaults]
IDX    CHAIN  CURVE NW ADDR_CLS         VB_PRV   VB_PUB   VB_WIF VB_ADDR  DFL_PATH NAME
0      -      secp  -  -                0488ade4 0488b21e -      -        0'/0/0   -

[bip-44]
IDX    CHAIN  CURVE NW ADDR_CLS         VB_PRV   VB_PUB   VB_WIF VB_ADDR  DFL_PATH NAME
0      BTC    x     m  P2PKH            x        x        80     00       x        Bitcoin
1      ---    x     T  P2PKH            04358394 043587cf ef     1d25     x        Testnet (all coins)
2      LTC    x     m  P2PKH            x        x        b0     spec     x        Litecoin
3      DOGE   x     m  P2PKH            02fac398 02facafd 9e     1e       x        Dogecoin
5      DASH   x     m  P2PKH            x        x        cc     4c       x        Dash
60     ETH    x     m  Eth              x        x        -      -        x        Ethereum
61     ETC    x     m  Eth              x        x        -      -        x        Ether Classic
74     ICX    x     m  Icx              x        x        -      -        x        ICON
77     XVG    x     m  P2PKH            x        x        9e     1e       x        Verge Currency
118    ATOM   x     m  Atom             x        x        -      h:stafi  x        Atom
128    XMR    x     m  Xmr              x        x        -      -        x        Monero
133    ZEC    x     m  P2PKH            x        x        80     1cb8     x        Zcash
144    XRP    x     m  Xrp              x        x        -      -        x        Ripple
145    BCH    x     m  BchP2PKH         x        x        80     spec     x        Bitcoin Cash
148    XLM    edw   m  Xlm              x        x        -      spec     0'       Stellar Lumens
165    XNO    blk   m  Nano             x        x        -      -        0'       Nano
194    EOS    x     m  Eos              x        x        -      -        x        EOS
195    TRX    x     m  Trx              x        x        -      -        x        Tron
236    BSV    x     m  P2PKH            x        x        80     00       x        BitcoinSV
283    ALGO   edw   m  Algo             x        x        -      -        0'/0'/0' Algorand
313    ZIL    x     m  Zil              x        x        -      -        x        Zilliqa
330    LUNA   x     m  Atom             x        x        -      h:terra  x        Terra
354    DOT    edw   m  SubstrateEd25519 x        x        -      spec     0'/0'/0' Polkadot
397    NEAR   edw   m  Near             x        x        -      -        0'       NEAR Protocol
429    ERG    x     T  ErgoP2PKH        04358394 043587cf -      spec     x        Ergo
434    KSM    edw   m  SubstrateEd25519 x        x        -      spec     0'/0'/0' Kusama
459    KAVA   x     m  Atom             x        x        -      h:kava   x        Kava
461    FIL    x     m  FilSecp256k1     x        x        -      -        x        Filecoin
494    BAND   x     m  Atom             x        x        -      h:band   x        Band
500    THETA  x     m  Eth              x        x        -      -        x        Theta
501    SOL    edw   m  Sol              x        x        -      -        0'       Solana
508    EGLD   edw   m  Egld             x        x        -      -        0'/0'/0' MultiversX
529    SCRT   x     m  Atom             x        x        -      h:secret x        Secret Network
567    NCG    x     m  Eth              x        x        -      -        x        Nine Chronicles
637    APTOS  edw   m  Aptos            x        x        -      -        0'/0'/0' Aptos
714    BNB    x     m  Atom             x        x        -      h:bnb    x        Binance
784    SUI    edw   m  Sui              x        x        -      -        0'/0'/0' Sui
818    VET    x     m  Eth              x        x        -      -        x        VeChain Token
888    NEO    nist  m  Neo              x        x        -      spec     x        NEO
931    RUNE   x     m  BechP2PKH        x        x        -      h:thor   x        THORChain
996    OKT    x     m  Okex             x        x        -      -        x        OKChain Token
1023   ONE    x     m  One              x        x        -      -        x        HARMONY-ONE (Legacy)
1024   ONT    nist  m  Neo              x        x        -      spec     x        Ontology
1729   XTZ    edw   m  Xtz              x        x        -      spec     0'/0'    Tezos
1815   ADA    khol  m  AdaByronIcarus   0f4331d4 x        -      spec     x        Cardano
9000   AVAX   x     m  AvaxXChain       x        x        -      -        x        Avalanche
52752  CELO   x     m  Eth              x        x        -      -        x        Celo
314159 PI     edw   m  Xlm              x        x        -      spec     0'       Pi Network

[bip-49]
IDX    CHAIN  CURVE NW ADDR_CLS         VB_PRV   VB_PUB   VB_WIF VB_ADDR  DFL_PATH NAME
0      BTC    x     m  P2SH             049d7878 049d7cb2 80     05       x        Bitcoin
1      ZEC    x     T  P2SH             044a4e28 044a5262 ef     1cba     x        Zcash TestNet
2      LTC    x     m  P2SH             049d7878 049d7cb2 b0     spec     x        Litecoin
3      DOGE   x     m  P2SH             02fac398 02facafd 9e     16       x        Dogecoin
5      DASH   x     m  P2SH             049d7878 049d7cb2 cc     10       x        Dash
133    ZEC    x     m  P2SH             049d7878 049d7cb2 80     1cbd     x        Zcash
145    XEC    x     m  BchP2SH          049d7878 049d7cb2 80     spec     x        eCash
236    BSV    x     m  P2SH             049d7878 049d7cb2 80     05       x        BitcoinSV

[bip-84]
IDX    CHAIN  CURVE NW ADDR_CLS         VB_PRV   VB_PUB   VB_WIF VB_ADDR  DFL_PATH NAME
0      BTC    x     m  P2WPKH           04b2430c 04b24746 80     h:bc     x        Bitcoin
1      LTC    x     T  P2WPKH           0436ef7d 0436f6e1 ef     h:tltc   x        Litecoin TestNet
2      LTC    x     m  P2WPKH           04b2430c 04b24746 b0     h:ltc    x        Litecoin

[bip-86]
IDX    CHAIN  CURVE NW ADDR_CLS         VB_PRV   VB_PUB   VB_WIF VB_ADDR  DFL_PATH NAME
0      BTC    x     m  P2TR             x        x        80     h:bc     x        Bitcoin
1      BTC    x     T  P2TR             04358394 043587cf ef     h:tb     x        Bitcoin TestNet

[bip-44-unsupported]
IDX    CHAIN    NAME
4      RDD    - Reddcoin
6      PPC    - Peercoin
7      NMC    - Namecoin
8      FTC    - Feathercoin
9      XCP    - Counterparty
10     BLK    - Blackcoin
11     NSR    - NuShares
12     NBT    - NuBits
13     MZC    - Mazacoin
14     VIA    - Viacoin
15     XCH    - ClearingHouse
16     RBY    - Rubycoin
17     GRS    - Groestlcoin
18     DGC    - Digitalcoin
19     CCN    - Cannacoin
20     DGB    - DigiByte
21     ---    - Open Assets
22     MONA   - Monacoin
23     CLAM   - Clams
24     XPM    - Primecoin
25     NEOS   - Neoscoin
26     JBS    - Jumbucks
27     ZRC    - ziftrCOIN
28     VTC    - Vertcoin
29     NXT    - NXT
30     BURST  - Burst
31     MUE    - MonetaryUnit
32     ZOOM   - Zoom
33     VASH   - Virtual Cash
34     CDN    - Canada eCoin
35     SDC    - ShadowCash
36     PKB    - ParkByte
37     PND    - Pandacoin
38     START  - StartCOIN
39     MOIN   - MOIN
40     EXP    - Expanse
41     EMC2   - Einsteinium
42     DCR    - Decred
43     XEM    - NEM
44     PART   - Particl
45     ARG    - Argentum (dead)
46     ---    - Libertas
47     ---    - Posw coin
48     SHR    - Shreeji
49     GCR    - Global Currency Reserve (GCRcoin)
50     NVC    - Novacoin
51     AC     - Asiacoin
52     BTCD   - BitcoinDark
53     DOPE   - Dopecoin
54     TPC    - Templecoin
55     AIB    - AIB
56     EDRC   - EDRCoin
57     SYS    - Syscoin
58     SLR    - Solarcoin
59     SMLY   - Smileycoin
62     PSB    - Pesobit
63     LDCN   - Landcoin (dead)
64     ---    - Open Chain
65     XBC    - Bitcoinplus
66     IOP    - Internet of People
67     NXS    - Nexus
68     INSN   - InsaneCoin
69     OK     - OKCash
70     BRIT   - BritCoin
71     CMP    - Compcoin
72     CRW    - Crown
73     BELA   - BelaCoin
75     FJC    - FujiCoin
76     MIX    - MIX
78     EFL    - Electronic Gulden
79     CLUB   - ClubCoin
80     RICHX  - RichCoin
81     POT    - Potcoin
82     QRK    - Quarkcoin
83     TRC    - Terracoin
84     GRC    - Gridcoin
85     AUR    - Auroracoin
86     IXC    - IXCoin
87     NLG    - Gulden
88     BITB   - BitBean
89     BTA    - Bata
90     XMY    - Myriadcoin
91     BSD    - BitSend
92     UNO    - Unobtanium
93     MTR    - MasterTrader
94     GB     - GoldBlocks
95     SHM    - Saham
96     CRX    - Chronos
97     BIQ    - Ubiquoin
98     EVO    - Evotion
99     STO    - SaveTheOcean
100    BIGUP  - BigUp
101    GAME   - GameCredits
102    DLC    - Dollarcoins
103    ZYD    - Zayedcoin
104    DBIC   - Dubaicoin
105    STRAT  - Stratis
106    SH     - Shilling
107    MARS   - MarsCoin
108    UBQ    - Ubiq
109    PTC    - Pesetacoin
110    NRO    - Neurocoin
111    ARK    - ARK
112    USC    - UltimateSecureCashMain
113    THC    - Hempcoin
114    LINX   - Linx
115    ECN    - Ecoin
116    DNR    - Denarius
117    PINK   - Pinkcoin
119    PIVX   - Pivx
120    FLASH  - Flashcoin
121    ZEN    - Zencash
122    PUT    - Putincoin
123    ZNY    - BitZeny
124    UNIFY  - Unify
125    XST    - StealthCoin
126    BRK    - Breakout Coin
127    VC     - Vcash
129    VOX    - Voxels
130    NAV    - NavCoin
131    FCT    - Factom Factoids
132    EC     - Factom Entry Credits
134    LSK    - Lisk
135    STEEM  - Steem
136    XZC    - ZCoin
137    RBTC   - RSK
138    ---    - Giftblock
139    RPT    - RealPointCoin
140    LBC    - LBRY Credits
141    KMD    - Komodo
142    BSQ    - bisq Token
143    RIC    - Riecoin
146    NEBL   - Neblio
147    ZCL    - ZClassic
149    NLC2   - NoLimitCoin2
150    WHL    - WhaleCoin
151    ERC    - EuropeCoin
152    DMD    - Diamond
153    BTM    - Bytom
154    BIO    - Biocoin
155    XWCC   - Whitecoin Classic
156    BTG    - Bitcoin Gold
157    BTC2X  - Bitcoin 2x
158    SSN    - SuperSkynet
159    TOA    - TOACoin
160    BTX    - Bitcore
161    ACC    - Adcoin
162    BCO    - Bridgecoin
163    ELLA   - Ellaism
164    PIRL   - Pirl
166    VIVO   - Vivo
167    FRST   - Firstcoin
168    HNC    - Helleniccoin
169    BUZZ   - BUZZ
170    MBRS   - Ember
171    HC     - Hcash
172    HTML   - HTMLCOIN
173    ODN    - Obsidian
174    ONX    - OnixCoin
175    RVN    - Ravencoin
176    GBX    - GoByte
177    BTCZ   - BitcoinZ
178    POA    - Poa
179    NYC    - NewYorkCoin
180    MXT    - MarteXcoin
181    WC     - Wincoin
182    MNX    - Minexcoin
183    BTCP   - Bitcoin Private
184    MUSIC  - Musicoin
185    BCA    - Bitcoin Atom
186    CRAVE  - Crave
187    STAK   - STRAKS
188    WBTC   - World Bitcoin
189    LCH    - LiteCash
190    EXCL   - ExclusiveCoin
191    ---    - Lynx
192    LCC    - LitecoinCash
193    XFE    - Feirm
196    KOBO   - Kobocoin
197    HUSH   - HUSH
198    BAN    - Banano
199    ETF    - ETF
200    OMNI   - Omni
201    BIFI   - BitcoinFile
202    UFO    - Uniform Fiscal Object
203    CNMC   - Cryptonodes
204    BCN    - Bytecoin
205    RIN    - Ringo
206    ATP    - Alaya
207    EVT    - everiToken
208    ATN    - ATN
209    BIS    - Bismuth
210    NEET   - NEETCOIN
211    BOPO   - BopoChain
212    OOT    - Utrum
213    ALIAS  - Alias
214    MONK   - Monkey Project
215    BOXY   - BoxyCoin
216    FLO    - Flo
217    MEC    - Megacoin
218    BTDX   - BitCloud
219    XAX    - Artax
220    ANON   - ANON
221    LTZ    - LitecoinZ
222    BITG   - Bitcoin Green
223    ICP    - Internet Computer (DFINITY)
224    SMART  - Smartcash
225    XUEZ   - XUEZ
226    HLM    - Helium
227    WEB    - Webchain
228    ACM    - Actinium
229    NOS    - NOS Stable Coins
230    BITC   - BitCash
231    HTH    - Help The Homeless Coin
232    TZC    - Trezarcoin
233    VAR    - Varda
234    IOV    - IOV
235    FIO    - FIO
237    DXN    - DEXON
238    QRL    - Quantum Resistant Ledger
239    PCX    - ChainX
240    LOKI   - Loki
241    ---    - Imagewallet
242    NIM    - Nimiq
243    SOV    - Sovereign Coin
244    JCT    - Jibital Coin
245    SLP    - Simple Ledger Protocol
246    EWT    - Energy Web
247    UC     - Ulord
248    EXOS   - EXOS
249    ECA    - Electra
250    SOOM   - Soom
251    XRD    - Redstone
252    FREE   - FreeCoin
253    NPW    - NewPowerCoin
254    BST    - BlockStamp
255    ---    - SmartHoldem
256    NANO   - Bitcoin Nano
257    BTCC   - Bitcoin Core
258    ---    - Zen Protocol
259    ZEST   - Zest
260    ABT    - ArcBlock
261    PION   - Pion
262    DT3    - DreamTeam3
263    ZBUX   - Zbux
264    KPL    - Kepler
265    TPAY   - TokenPay
266    ZILLA  - ChainZilla
267    ANK    - Anker
268    BCC    - BCChain
269    HPB    - HPB
270    ONE    - ONE
271    SBC    - SBC
272    IPC    - IPChain
273    DMTC   - Dominantchain
274    OGC    - Onegram
275    SHIT   - Shitcoin
276    ANDES  - Andescoin
277    AREPA  - Arepacoin
278    BOLI   - Bolivarcoin
279    RIL    - Rilcoin
280    HTR    - Hathor Network
281    ACME   - Accumulate
282    BRAVO  - BRAVO
284    BZX    - Bitcoinzero
285    GXX    - GravityCoin
286    HEAT   - HEAT
287    XDN    - DigitalNote
288    FSN    - FUSION
289    CPC    - Capricoin
290    BOLD   - Bold
291    IOST   - IOST
292    TKEY   - Tkeycoin
293    USE    - Usechain
294    BCZ    - BitcoinCZ
295    IOC    - Iocoin
296    ASF    - Asofe
297    MASS   - MASS
298    FAIR   - FairCoin
299    NUKO   - Nekonium
300    GNX    - Genaro Network
301    DIVI   - Divi Project
302    CMT    - Community
303    EUNO   - EUNO
304    IOTX   - IoTeX
305    ONION  - DeepOnion
306    8BIT   - 8Bit
307    ATC    - AToken Coin
308    BTS    - Bitshares
309    CKB    - Nervos CKB
310    UGAS   - Ultrain
311    ADS    - Adshares
312    ARA    - Aura
314    MOAC   - MOAC
315    SWTC   - SWTC
316    VNSC   - vnscoin
317    PLUG   - Pl^g
318    MAN    - Matrix AI Network
319    ECC    - ECCoin
320    RPD    - Rapids
321    RAP    - Rapture
322    GARD   - Hashgard
323    ZER    - Zero
324    EBST   - eBoost
325    SHARD  - Shard
326    MRX    - Metrix Coin
327    CMM    - Commercium
328    BLOCK  - Blocknet
329    AUDAX  - AUDAX
331    ZPM    - zPrime
332    KUVA   - Kuva Utility Note
333    MEM    - MemCoin
334    CS     - Credits
335    SWIFT  - SwiftCash
336    FIX    - FIX
337    CPC    - CPChain
338    VGO    - VirtualGoodsToken
339    DVT    - DeVault
340    N8V    - N8VCoin
341    MTNS   - OmotenashiCoin
342    BLAST  - BLAST
343    DCT    - DECENT
344    AUX    - Auxilium
345    USDP   - USDP
346    HTDF   - HTDF
347    YEC    - Ycash
348    QLC    - QLC Chain
349    TEA    - Icetea Blockchain
350    ARW    - ArrowChain
351    MDM    - Medium
352    CYB    - Cybex
353    LTO    - LTO Network
355    AEON   - Aeon
356    RES    - Resistance
357    AYA    - Aryacoin
358    DAPS   - Dapscoin
359    CSC    - CasinoCoin
360    VSYS   - V Systems
361    NOLLAR - Nollar
362    XNOS   - NOS
363    CPU    - CPUchain
364    LAMB   - Lambda Storage Chain
365    VCT    - ValueCyber
366    CZR    - Canonchain
367    ABBC   - ABBC
368    HET    - HET
369    XAS    - Asch
370    VDL    - Vidulum
371    MED    - MediBloc
372    ZVC    - ZVChain
373    VESTX  - Vestx
374    DBT    - DarkBit
375    SEOS   - SuperEOS
376    MXW    - Maxonrow
377    ZNZ    - ZENZO
378    XCX    - XChain
379    SOX    - SonicX
380    NYZO   - Nyzo
381    ULC    - ULCoin
382    RYO    - Ryo Currency
383    KAL    - Kaleidochain
384    XSN    - Stakenet
385    DOGEC  - DogeCash
386    BMV    - Bitcoin Matteo's Vision
387    QBC    - Quebecoin
388    IMG    - ImageCoin
389    QOS    - QOS
390    PKT    - PKT
391    LHD    - LitecoinHD
392    CENNZ  - CENNZnet
393    HSN    - Hyper Speed Network
394    CRO    - Crypto Chain
395    UMBRU  - Umbru
396    EVER   - Everscale
398    XPC    - XPChain
399    ZOC    - 01coin
400    NIX    - NIX
401    UC     - Utopiacoin
402    GALI   - Galilel
403    OLT    - Oneledger
404    XBI    - XBI
405    DONU   - DONU
406    EARTHS - Earths
407    HDD    - HDDCash
408    SUGAR  - Sugarchain
409    AILE   - AileCoin
410    TENT   - TENT
411    TAN    - Tangerine Network
412    AIN    - AIN
413    MSR    - Masari
414    SUMO   - Sumokoin
415    ETN    - Electroneum
416    BYTZ   - BYTZ
417    WOW    - Wownero
418    XTNC   - XtendCash
419    LTHN   - Lethean
420    NODE   - NodeHost
421    AGM    - Argoneum
422    CCX    - Conceal Network
423    TNET   - Title Network
424    TELOS  - TelosCoin
425    AION   - Aion
426    BC     - Bitcoin Confidential
427    KTV    - KmushiCoin
428    ZCR    - ZCore
430    PESO   - Criptopeso
431    BTC2   - Bitcoin 2
432    XRPHD  - XRPHD
433    WE     - WE Coin
435    PCN    - Peepcoin
436    NCH    - NetCloth
437    ICU    - CHIPO
438    FNSA   - FINSCHIA
439    DTP    - DeVault Token Protocol
440    BTCR   - Bitcoin Royale
441    AERGO  - AERGO
442    XTH    - Dothereum
443    LV     - Lava
444    PHR    - Phore
445    VITAE  - Vitae
446    COCOS  - Cocos-BCX
447    DIN    - Dinero
448    SPL    - Simplicity
449    YCE    - MYCE
450    XLR    - Solaris
451    KTS    - Klimatas
452    DGLD   - DGLD
453    XNS    - Insolar
454    EM     - EMPOW
455    SHN    - ShineBlocks
456    SEELE  - Seele
457    AE     - æternity
458    ODX    - ObsidianX
460    GLEEC  - GLEEC
462    RUTA   - Rutanio
463    CSDT   - CSDT
464    ETI    - EtherInc
465    ZSLP   - Zclassic Simple Ledger Protocol
466    ERE    - EtherCore
467    DX     - DxChain Token
468    CPS    - Capricoin+
469    BTH    - Bithereum
470    MESG   - MESG
471    FIMK   - FIMK
472    AR     - Arweave
473    OGO    - Origo
474    ROSE   - Oasis Network
475    BARE   - BARE Network
476    GLEEC  - GleecBTC
477    CLR    - Color Coin
478    RNG    - Ring
479    OLO    - Tool Global
480    PEXA   - Pexa
481    MOON   - Mooncoin
482    OCEAN  - Ocean Protocol
483    BNT    - Bluzelle Native
484    AMO    - AMO Blockchain
485    FCH    - FreeCash
486    LAT    - PlatON
487    COIN   - Bitcoin Bank
488    VEO    - Amoveo
489    CCA    - Counos Coin
490    GFN    - Graphene
491    BIP    - Minter Network
492    KPG    - Kunpeng Network
493    FIN    - FINL Chain
495    DROP   - Dropil
496    BHT    - Bluehelix Chain
497    LYRA   - Scrypta
498    CS     - Credits
499    RUPX   - Rupaya
502    THT    - ThoughtAI
503    CFX    - Conflux
504    KUMA   - Kumacoin
505    HASH   - Provenance
506    CSPR   - Casper
507    EARTH  - EARTH
509    CHI    - Xaya
510    KOTO   - Koto
511    OTC    - θ
512    XRD    - Radiant
513    SEELEN - Seele-N
514    AETH   - AETH
515    DNA    - Idena
516    VEE    - Virtual Economy Era
517    SIERRA - SierraCoin
518    LET    - Linkeye
519    BSC    - Bitcoin Smart Contract
520    BTCV   - BitcoinVIP
521    ABA    - Dabacus
522    SCC    - StakeCubeCoin
523    EDG    - Edgeware
524    AMS    - AmsterdamCoin
525    GOSS   - GOSSIP Coin
526    BU     - BUMO
527    GRAM   - GRAM
528    YAP    - Yapstone
530    NOVO   - Novo
531    GHOST  - Ghost
532    HST    - HST
533    PRJ    - ProjectCoin
534    YOU    - YOUChain
535    XHV    - Haven Protocol
536    BYND   - Beyondcoin
537    JOYS   - Joys Digital
538    VAL    - Valorbit
539    FLOW   - Flow
540    SMESH  - Spacemesh Coin
541    SCDO   - SCDO
542    IQS    - IQ-Cash
543    BIND   - Compendia
544    COINEVO - Coinevo
545    SCRIBE - Scribe
546    HYN    - Hyperion
547    BHP    - BHP
548    BBC    - BigBang Core
549    MKF    - MarketFinance
550    XDC    - XinFin
551    STR    - Straightedge
552    SUM    - Sumcoin
553    HBC    - HuobiChain
554    ---    - reserved
555    BCS    - Bitcoin Smart
556    KTS    - Kratos
557    LKR    - Lkrcoin
558    TAO    - Tao
559    XWC    - Whitecoin
560    DEAL   - DEAL
561    NTY    - Nexty
562    TOP    - TOP NetWork
563    ---    - reserved
564    AG     - Agoric
565    CICO   - Coinicles
566    IRIS   - Irisnet
568    LRG    - Large Coin
569    SERO   - Super Zero Protocol
570    BDX    - Beldex
571    CCXX   - Counos X
572    SLS    - Saluscoin
573    SRM    - Serum
574    ---    - reserved
575    VIVT   - VIDT Datalink
576    BPS    - BitcoinPoS
577    NKN    - NKN
578    ICL    - ILCOIN
579    BONO   - Bonorum
580    PLC    - PLATINCOIN
581    DUN    - Dune
582    DMCH   - Darmacash
583    CTC    - Creditcoin
584    KELP   - Haidai Network
585    GBCR   - GoldBCR
586    XDAG   - XDAG
587    PRV    - Incognito Privacy
588    SCAP   - SafeCapital
589    TFUEL  - Theta Fuel
590    GTM    - Gentarium
591    RNL    - RentalChain
592    GRIN   - Grin
593    MWC    - MimbleWimbleCoin
594    DOCK   - Dock
595    POLYX  - Polymesh
596    DIVER  - Divergenti
597    XEP    - Electra Protocol
598    APN    - Apron
599    TFC    - Turbo File Coin
600    UTE    - Unit-e
601    MTC    - Metacoin
602    NC     - NobodyCash
603    XINY   - Xinyuehu
604    DYN    - Dynamo
605    BUFS   - Buffer
606    STOS   - Stratos
607    TON    - TON
608    TAFT   - TAFT
609    HYDRA  - HYDRA
610    NOR    - Noir
611    ---    - Manta Network Private Asset
612    ---    - Calamari Network Private Asset
613    WCN    - Widecoin
614    OPT    - Optimistic Ethereum
615    PSWAP  - PolkaSwap
616    VAL    - Validator
617    XOR    - Sora
618    SSP    - SmartShare
619    DEI    - DeimosX
620    ---    - reserved
621    ZERO   - Singularity
622    ALPHA  - AlphaDAO
623    BDECO  - BDCashProtocol Ecosystem
624    NOBL   - Nobility
625    EAST   - Eastcoin
626    KDA    - Kadena
627    SOUL   - Phantasma
628    LORE   - Gitopia
629    FNR    - Fincor
630    NEXUS  - Nexus
631    QTZ    - Quartz
632    MAS    - Massa
633    CALL   - Callchain
634    VAL    - Validity
635    POKT   - Pocket Network
636    EMIT   - EMIT
638    ADON   - ADON
639    BTSG   - BitSong
640    LFC    - Leofcoin
641    KCS    - KuCoin Shares
642    KCC    - KuCoin Community Chain
643    AZERO  - Aleph Zero
644    TREE   - Tree
645    LX     - Lynx
646    XLN    - Lunarium
647    CIC    - CIC Chain
648    ZRB    - Zarb
649    ---    - reserved
650    UCO    - Archethic
651    SFX    - Safex Cash
652    SFT    - Safex Token
653    WSFX   - Wrapped Safex Cash
654    USDG   - US Digital Gold
655    WMP    - WAMP
656    EKTA   - Ekta
657    YDA    - YadaCoin
659    KOIN   - Koinos
660    PIRATE - PirateCash
661    UNQ    - Unique
663    SFRX   - EtherGem Sapphire
666    ACT    - Achain
667    PRKL   - Perkle
668    SSC    - SelfSell
669    GC     - GateChain
670    PLGR   - Pledger
671    MPLGR  - Pledger
672    KNOX   - Knox
673    ZED    - ZED
674    CNDL   - Candle
675    WLKR   - Walker Crypto Innovation Index
676    WLKRR  - Walker
677    YUNGE  - Yunge
678    Voken  - Voken
679    APL    - Apollo
680    Evrynet - Evrynet
681    NENG   - Nengcoin
682    CHTA   - Cheetahcoin
683    ALEO   - Aleo Network
685    OAS    - Oasys
686    KAR    - Karura Network
688    CET    - CoinEx Chain
690    KLV    - KleverChain
694    VTBC   - VTB Community
698    VEIL   - Veil
699    GTB    - GotaBit
700    XDAI   - xDai
701    COM    - Commercio
702    CCC    - Commercio Cash Credit
707    MCOIN  - Moneta Coin
710    FURY   - Highbury
711    CHC    - Chaincoin
712    SERF   - Serfnet
713    XTL    - Katal Chain
715    SIN    - Sinovate
716    DLN    - Delion
717    BONTE  - Bontecoin
718    PEER   - Peer
719    ZET    - Zetacoin
720    ABY    - Artbyte
721    PGX    - Mirai Chain
722    IL8P   - InfiniLooP
724    XVC    - Vanillacash
725    MCX    - MultiCash
727    BLU    - BluCrates
730    HEALIOS - Tenacity
731    BMK    - Bitmark
734    DENTX  - DENTNet
737    ATOP   - Financial Blockchain
747    CFG    - Centrifuge
750    XPRT   - Persistence
753    ---    - Age X25519 Encryption
754    ---    - Age NIST Encryption
757    HONEY  - HoneyWood
768    BALLZ  - Ballzcoin
770    COSA   - Cosanta
771    BR     - BR
775    PLSR   - Pulsar Coin
776    KEY    - Keymaker Coin
777    BTW    - Bitcoin World
780    PLCUC  - PLC Ultima Classic
781    PLCUX  - PLC Ultima X
782    PLCU   - PLC Ultima
783    SMARTBC - SMART Blockchain
786    UIDD   - UIDD
787    ACA    - Acala
788    BNC    - Bifrost
789    TAU    - Lamden
799    PDEX   - Polkadex
800    BEET   - Beetle Coin
801    DST    - DSTRA
802    CY     - Cyberyen
804    ZKS    - zkSync
808    QVT    - Qvolta
809    SDN    - Shiden Network
810    ASTR   - Astar Network
811    ---    - reserved
813    MEER   - Qitmeer
819    REEF   - Reef
820    CLO    - Callisto
822    BDB    - BigchainDB
827    ACE    - Endurance
828    CCN    - ComputeCoin
829    BBA    - BBACHAIN
831    CRUZ   - cruzbit
832    SAPP   - Sapphire
833    777    - Jackpot
834    KYAN   - Kyanite
835    AZR    - Azzure
836    CFL    - CryptoFlow
837    DASHD  - Dash Diamond
838    TRTT   - Trittium
839    UCR    - Ultra Clear
840    PNY    - Peony
841    BECN   - Beacon
842    MONK   - Monk
843    SAGA   - CryptoSaga
844    SUV    - Suvereno
845    ESK    - EskaCoin
846    OWO    - OneWorld Coin
847    PEPS   - PEPS Coin
848    BIR    - Birake
849    MOBIC  - MobilityCoin
850    FLS    - Flits
852    DSM    - Desmos
853    PRCY   - PRCY Coin
858    HVH    - HAVAH
866    MOB    - MobileCoin
868    IF     - Infinitefuture
877    NAM    - Namada
878    SCR    - Scorum Network
880    LUM    - Lum Network
883    ZBC    - ZooBC
886    ADF    - AD Token
889    TOMO   - TOMO
890    XSEL   - Seln
896    LKSC   - LKSCoin
898    AS     - Assetchain
899    XEC    - eCash
900    LMO    - Lumeneo
901    NXT    - NxtMeta
904    HNT    - Helium
907    FIS    - StaFi
909    SGE    - Saage
911    GERT   - Gert
913    VARA   - Vara Network
916    META   - Metadium
917    FRA    - Findora
919    CCD    - Concordium
921    AVN    - Avian Network
925    DIP    - Dipper Network
928    GHM    - HermitMatrixNetwork
941    ---    - reserved
945    UNLOCK - Jasiri protocol
955    LTP    - LifetionCoin
958    ---    - KickSoccer
960    VKAX   - Vkax
966    MATIC  - Matic
968    UNW    - UNW
970    TWINS  - TWINS
977    TLOS   - Telos
981    TAFECO - Taf ECO Chain
985    AU     - Autonomy
987    VCG    - VipCoin
988    XAZAB  - Xazab core
989    AIOZ   - AIOZ
990    CORE   - Coreum
991    PEC    - Phoenix
992    UNT    - Unit
993    XRB    - X Currency
994    QUAI   - Quai Network
995    CAPS   - Ternoa
997    SUM    - Solidum
998    LBTC   - Lightning Bitcoin
999    BCD    - Bitcoin Diamond
1000   BTN    - Bitcoin New
1001   TT     - ThunderCore
1002   BKT    - BanKitt
1003   NODL   - Nodle
1004   PCOIN  - PCOIN
1005   TAO    - Bittensor
1006   HSK    - HashKey Chain
1007   FTM    - Fantom
1008   RPG    - RPG
1009   LAKE   - iconLake
1010   HT     - Huobi ECO Chain
1011   ELV    - Eluvio
1012   JOC    - Japan Open Chain
1013   BIC    - Beincrypto
1016   ---    - reserved
1020   EVC    - Evrice
1022   XRD    - Radix DLT
1025   CZZ    - Classzz
1026   KEX    - Kira Exchange Token
1027   MCM    - Mochimo
1028   PLS    - Pulse Coin
1032   BTCR   - BTCR
1042   MFID   - Moonfish ID
1111   BBC    - Big Bitcoin
1116   CORE   - Core
1120   RISE   - RISE
1122   CMT    - CyberMiles Token
1128   ETSC   - Ethereum Social
1129   DFI    - DeFiChain
1130   DFI    - DeFiChain EVM Network
1137   $DAG   - Constellation Labs
1145   CDY    - Bitcoin Candy
1155   ENJ    - Enjin Coin
1170   HOO    - Hoo Smart Chain
1234   ALPH   - Alephium
1236   ---    - Masca
1237   ---    - Nostr
1280   ---    - Kudos Setler
1284   GLMR   - Moonbeam
1285   MOVR   - Moonriver
1298   WPC    - Wpc
1308   WEI    - WEI
1337   DFC    - Defcoin
1348   ISLM   - IslamicCoin
1397   HYC    - Hycon
1410   TENTSLP - TENT Simple Ledger Protocol
1510   XSC    - XT Smart Chain
1512   AAC    - Double-A Chain
1524   ---    - Taler
1533   BEAM   - Beam
1551   SDK    - Sovereign SDK
1555   APC    - Apc Chain
1616   ELF    - AELF
1618   AUDL   - AUDL
1620   ATH    - Atheios
1627   LUME   - Lume Web
1642   NEW    - Newton
1657   BTA    - Btachain
1668   NEOX   - Neoxa
1669   MEWC   - Meowcoin
1688   BCX    - BitcoinX
1776   LBTC   - Liquid BTC
1777   BBP    - Biblepay
1784   JPYS   - JPY Stablecoin
1789   VEGA   - Vega Protocol
1818   CUBE   - Cube Chain Native Token
1856   TES    - Teslacoin
1888   ZTX    - Zetrix
1899   XEC    - eCash token
1901   CLC    - Classica
1907   BITCI  - Bitcicoin
1919   VIPS   - VIPSTARCOIN
1926   CITY   - City Coin
1955   XX     - xx coin
1977   XMX    - Xuma
1984   TRTL   - TurtleCoin
1985   SLRT   - Solarti Chain
1986   QTH    - Qing Tong Horizon
1987   EGEM   - EtherGem
1988   MIRA   - Mira Chain
1989   HODL   - HOdlcoin
1990   PHL    - Placeholders
1991   SC     - Sia
1996   MYT    - Mineyourtime
1997   POLIS  - Polis
1998   XMCC   - Monoeci
1999   COLX   - ColossusXT
2000   GIN    - GinCoin
2001   MNP    - MNPCoin
2002   MLN    - Miraland
2017   KIN    - Kin
2018   EOSC   - EOSClassic
2019   GBT    - GoldBean Token
2020   PKC    - PKC
2021   SKT    - Sukhavati
2022   XHT    - Xinghuo Token
2023   COC    - Chat On Chain
2024   USBC   - Universal Ledger USBC
2046   ANY    - Any
2048   MCASH  - MCashChain
2049   TRUE   - TrueChain
2050   MOVO   - Movo Smart Chain
2086   KILT   - KILT Spiritnet
2109   SAMA   - Exosama Network
2112   IoTE   - IoTE
2125   BAY    - BitBay
2137   XRG    - Ergon
2182   CHZ    - Chiliz
2199   SAMA   - Moonsama Network
2221   ASK    - ASK
2222   CWEB   - Coinweb
2285   ---    - Qiyi Chain
2301   QTUM   - QTUM
2302   ETP    - Metaverse
2303   GXC    - GXChain
2304   CRP    - CranePay
2305   ELA    - Elastos
2338   SNOW   - Snowblossom
2365   XIN    - Mixin
2500   NEXI   - Nexi
2570   AOA    - Aurora
2718   NAS    - Nebulas
2894   REOSC  - REOSC Ecosystem
2941   BND    - Blocknode
3000   SM     - Stealth Message
3003   LUX    - LUX
3030   HBAR   - Hedera HBAR
3077   COS    - Contentos
3276   CCC    - CodeChain
3333   SXP    - Solar
3377   ROI    - ROIcoin
3381   DYN    - Dynamic
3383   SEQ    - Sequence
3552   DEO    - Destocoin
3564   DST    - DeStream
3601   CY     - Cybits
3757   MPC    - Partisia Blockchain
4040   FC8    - FCH Network
4096   YEE    - YeeCo
4218   IOTA   - IOTA
4219   SMR    - Shimmer
4242   AXE    - Axe
4343   XYM    - Symbol
4444   C4E    - Chain4Energy
4919   XVM    - Venidium
4999   BXN    - BlackFort Exchange Network
5006   SBC    - Senior Blockchain
5248   FIC    - FIC
5353   HNS    - Handshake
5404   ISK    - ISKRA
5467   ALTME  - ALTME
5555   FUND   - Unification
5757   STX    - Stacks
5895   VOW    - VowChain VOW
5920   SLU    - SILUBIUM
6060   GO     - GoChain GO
6174   MOI    - My Own Internet
6532   UM     - Penumbra
6599   RSC    - Royal Sports City
6666   BPA    - Bitcoin Pizza
6688   SAFE   - SAFE
6779   COTI   - COTI
6969   ROGER  - TheHolyrogerCoin
7027   ELLA   - Ella the heart
7028   AA     - Arthera
7091   TOPL   - Topl
7331   KLY    - KLYNTAR
7341   SHFT   - Shyft
7518   MEV    - MEVerse
7576   ADIL   - ADIL Chain
7777   BTV    - Bitvote
8000   SKY    - Skycoin
8080   ---    - DSRV
8181   BOC    - BeOne Chain
8192   PAC    - pacprotocol
8217   KLAY   - KLAY
8339   BTQ    - BitcoinQuark
8444   XCH    - Chia
8520   ---    - reserved
8680   PLMNT  - Planetmint
8866   GGX    - Golden Gate
8886   GGXT   - Golden Gate Sydney
8888   SBTC   - Super Bitcoin
8964   NULS   - NULS
8997   BBC    - Babacoin
8998   JGC    - JagoanCoin
8999   BTP    - Bitcoin Pay
9001   ARB1   - Arbitrum
9002   BOBA   - Boba
9003   LOOP   - Loopring
9004   STRK   - StarkNet
9005   AVAXC  - Avalanche C-Chain
9006   BSC    - Binance Smart Chain
9797   NRG    - Energi
9888   BTF    - Bitcoin Faith
9999   GOD    - Bitcoin God
10000  FO     - FIBOS
10111  DHP    - dHealth
10226  RTM    - Raptoreum
10291  XRC    - XRhodium
10507  NUM    - Numbers Protocol
10605  XPI    - Lotus
11111  ESS    - Essentia One
11742  VARCH  - InvArch
11743  TNKR   - Tinkernet
12345  IPOS   - IPOS
12586  MINA   - Mina
13107  BTY    - BitYuan
13108  YCC    - Yuan Chain Coin
14001  WAX    - Worldwide Asset Exchange
15845  SDGO   - SanDeGo
16181  XTX    - Totem Live Network
16754  ARDR   - Ardor
18000  MTR    - Meter
19165  SAFE   - Safecoin
19167  FLUX   - Flux
19169  RITO   - Ritocoin
19788  ML     - Mintlayer
20036  XND    - ndau
21004  C4EI   - c4ei
21888  PAC    - Pactus
22504  PWR    - PWRcoin
23000  EPIC   - Epic Cash
25252  BELL   - Bellcoin
25718  CHX    - Own
29223  NEXA   - Nexa
30001  ---    - reserved
31102  ESN    - EtherSocial Network
31337  ---    - ThePower
33416  TEO    - Trust Eth reOrigin
33878  BTCS   - Bitcoin Stake
34952  BTT    - ByteTrade
37992  FXTC   - FixedTradeCoin
39321  AMA    - Amabig
42069  FACT   - FACT0RN
43028  AXIV   - AXIV
49262  EVE    - evan
49344  STASH  - STASH
61616  TH     - TianHe
65536  KETH   - Krypton World
69420  GRLC   - Garlicoin
70007  GWL    - Gewel
77777  ZYN    - Wethio
88888  RYO    - c0ban
99999  WICC   - Waykichain
100500 HOME   - HomeCoin
101010 STC    - Starcoin
105105 STRAX  - Strax
111111 KAS    - Kaspa
161803 APTA   - Bloqs4Good
200625 AKA    - Akroma
200665 GENOM  - GENOM
246529 ATS    - ARTIS sigma1
261131 ZAMA   - Zama
333332 VALUE  - Value Chain
333333 3333   - Pi Value Consensus
424242 X42    - x42
534352 SCR    - Scroll
666666 VITE   - Vite
888888 SEA    - Second Exchange Alliance
999999 WTC    - WaltonChain
1048576 AMAX   - Armonia Meta Chain
1171337 ILT    - iOlite
1313114 ETHO   - Etho Protocol
1313500 XERO   - Xerom
1712144 LAX    - LAPO
3924011 EPK    - EPIK Protocol
4741444 HYD    - Hydra Token
5249353 BCO    - BitcoinOre
5249354 BHD    - BitcoinHD
5264462 PTN    - PalletOne
5655640 VLX    - Velas
5718350 WAN    - Wanchain
5741564 WAVES  - Waves
5741565 WEST   - Waves Enterprise
6382179 ABC    - Abcmint
6517357 CRM    - Creamcoin
7171666 BROCK  - Bitrock
7562605 SEM    - Semux
7567736 ION    - ION
7777777 FCT    - FirmaChain
7825266 WGR    - WGR
7825267 OBSR   - OBServer
8163271 AFS    - ANFS
11259375 LBR    - 0L
15118976 XDS    - XDS
61717561 AQUA   - Aquachain
88888888 HATCH  - Hatch
91927009 kUSD   - kUSD
99999996 GENS   - GENS
99999997 EQ     - EQ
99999998 FLUID  - Fluid Chains
99999999 QKC    - QuarkChain
608589380 FVDC   - ForumCoin
1179993420 ---    - Fuel

"""



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
mmgen-tool:  Perform various MMGen- and cryptocoin-related operations.
             Part of the MMGen suite
"""

import sys, os, importlib
from .cfg import gc, Config
from .util import msg, Msg, die, capfirst, suf, async_run

opts_data = {
	'filter_codes': ['-'],
	'text': {
		'desc':    f'Perform various {gc.proj_name}- and cryptocoin-related operations',
		'usage':   '[opts] <command> <command args>',
		'options': """
			-- -d, --outdir=       d  Specify an alternate directory 'd' for output
			-- -h, --help             Print this help message
			-- --, --longhelp         Print help message for long (global) options
			-- -e, --echo-passphrase  Echo passphrase or mnemonic to screen upon entry
			-- -k, --use-internal-keccak-module Force use of the internal keccak module
			-- -K, --keygen-backend=n Use backend 'n' for public key generation.  Options
			+                         for {coin_id}: {kgs}
			-- -l, --list             List available commands
			-- -p, --hash-preset= p   Use the scrypt hash parameters defined by preset 'p'
			+                         for password hashing (default: '{gc.dfl_hash_preset}')
			-- -P, --passwd-file= f   Get passphrase from file 'f'.
			-- -q, --quiet            Produce quieter output
			-- -r, --usr-randchars=n  Get 'n' characters of additional randomness from
			+                         user (min={cfg.min_urandchars}, max={cfg.max_urandchars})
			-- -t, --type=t           Specify address type (valid choices: 'legacy',
			+                         'compressed', 'segwit', 'bech32', 'zcash_z')
			-- -v, --verbose          Produce more verbose output
			-- -x, --proxy=P          Proxy HTTP connections via SOCKS5h proxy ‘P’ (host:port).
			+                         Use special value ‘env’ to honor *_PROXY environment
			+                         vars instead.
			e- -X, --cached-balances  Use cached balances
			-- -y, --yes              Answer 'yes' to prompts, suppress non-essential output
""",
	'notes': """

                               COMMANDS

{ch}
Type ‘{pn} help <command>’ for help on a particular command
"""
	},
	'code': {
		'options': lambda cfg, s, help_notes: s.format(
			kgs     = help_notes('keygen_backends'),
			coin_id = help_notes('coin_id'),
			cfg     = cfg,
			gc      = gc,
		),
		'notes': lambda cfg, s, help_notes: s.format(
			ch      = help_notes('tool_help'),
			pn      = gc.prog_name)
	}
}

# NB: Command groups and commands are displayed on the help screen in the following order,
# so keep the command names sorted
mods = {
	'help': (
		'help',
		'usage',
	),
	'util': (
		'b32tohex',
		'b58chktohex',
		'b58tobytes',
		'b58tohex',
		'b6dtohex',
		'bytespec',
		'bytestob58',
		'hash160',
		'hash256',
		'hexdump',
		'hexlify',
		'hexreverse',
		'hextob32',
		'hextob58',
		'hextob58chk',
		'hextob6d',
		'id6',
		'id8',
		'randb58',
		'randhex',
		'str2id6',
		'to_bytespec',
		'unhexdump',
		'unhexlify',
	),
	'coin': (
		'addr2pubhash',
		'addr2scriptpubkey',
		'eth_checksummed_addr',
		'hex2wif',
		'privhex2addr',
		'privhex2pubhex',
		'pubhash2addr',
		'pubhex2addr',
		'pubhex2redeem_script',
		'privhex2pair',
		'randpair',
		'randwif',
		'redeem_script2addr',
		'scriptpubkey2addr',
		'wif2addr',
		'wif2hex',
		'wif2redeem_script',
		'wif2segwit_pair',
	),
	'mnemonic': (
		'hex2mn',
		'mn2hex',
		'mn2hex_interactive',
		'mn_printlist',
		'mn_rand128',
		'mn_rand192',
		'mn_rand256',
		'mn_stats',
	),
	'file': (
		'addrfile_chksum',
		'keyaddrfile_chksum',
		'viewkeyaddrfile_chksum',
		'passwdfile_chksum',
		'txview',
	),
	'filecrypt': (
		'decrypt',
		'encrypt',
	),
	'fileutil': (
		'decrypt_keystore',
		'decrypt_geth_keystore',
		'find_incog_data',
		'rand2file',
	),
	'wallet': (
		'gen_addr',
		'gen_key',
		'get_subseed',
		'get_subseed_by_seed_id',
		'list_shares',
		'list_subseeds',
	),
	'rpc': (
		'add_label',
		'daemon_version',
		'getbalance',
		'listaddress',
		'listaddresses',
		'remove_address',
		'remove_label',
		'rescan_address',
		'rescan_blockchain',
		'resolve_address',
		'twexport',
		'twimport',
		'twview',
		'txhist',
	),
}

def get_cmds():
	return [cmd for mod, cmds in mods.items() if mod != 'help' for cmd in cmds]

def create_call_sig(cmd, cls, *, as_string=False):

	m = getattr(cls, cmd)

	if 'varargs_call_sig' in m.__code__.co_varnames: # hack
		flag = 'VAR_ARGS'
		va = m.__defaults__[0]
		args, dfls, ann = va['args'], va['dfls'], va['annots']
	else:
		flag = None
		c = m.__code__
		args = c.co_varnames[1:c.co_argcount + c.co_posonlyargcount + c.co_kwonlyargcount]
		dfls = (
			(m.__defaults__ or ()) +
			tuple(m.__kwdefaults__[k] for k in args if k in (m.__kwdefaults__ or ())))
		ann  = m.__annotations__

	nargs = len(args) - len(dfls)
	dfl_types = tuple(
		ann[a] if a in ann and isinstance(ann[a], type) else type(dfls[i])
			for i, a in enumerate(args[nargs:]))

	if as_string:
		get_type_from_ann = lambda x: 'str or STDIN' if ann[x] == 'sstr' else ann[x].__name__
		return ' '.join(
			[f'{a} [{get_type_from_ann(a)}]' for a in args[:nargs]] +
			[f'{a} [{dfl_types[n].__name__}={dfls[n]!r}]' for n, a in enumerate(args[nargs:])])
	else:
		get_type_from_ann = lambda x: 'str' if ann[x] == 'sstr' else ann[x].__name__
		return (
			[(a, get_type_from_ann(a)) for a in args[:nargs]],            # c_args
			{a:dfls[n] for n, a in enumerate(args[nargs:])},              # c_kwargs
			{a:dfl_types[n] for n, a in enumerate(args[nargs:])},         # c_kwargs_types
			('STDIN_OK' if nargs and ann[args[0]] == 'sstr' else flag),  # flag
			ann)                                                        # ann

def process_args(cmd, cmd_args, cls):
	c_args, c_kwargs, c_kwargs_types, flag, _ = create_call_sig(cmd, cls)
	have_stdin_input = False

	def usage_die(s):
		msg(s)
		from .tool.help import usage
		usage(cmd)

	if flag != 'VAR_ARGS':
		if len(cmd_args) < len(c_args):
			usage_die(f'Command requires exactly {len(c_args)} non-keyword argument{suf(c_args)}')

		u_args = cmd_args[:len(c_args)]

		# If we're reading from a pipe, replace '-' with output of previous command
		if flag == 'STDIN_OK' and u_args and u_args[0] == '-':
			if sys.stdin.isatty():
				die('BadFilename', "Standard input is a TTY.  Can't use '-' as a filename")
			else:
				from .util2 import parse_bytespec
				max_dlen_spec = '10kB' # limit input to 10KB for now
				max_dlen = parse_bytespec(max_dlen_spec)
				u_args[0] = os.read(0, max_dlen)
				have_stdin_input = True
				if len(u_args[0]) >= max_dlen:
					die(2, f'Maximum data input for this command is {max_dlen_spec}')
				if not u_args[0]:
					die(2, f'{cmd}: ERROR: no output from previous command in pipe')

	u_nkwargs = len(cmd_args) - len(c_args)
	u_kwargs = {}
	if flag == 'VAR_ARGS':
		cmd_args = ['dummy_arg'] + cmd_args
		t = [a.split('=', 1) for a in cmd_args if '=' in a]
		tk = [a[0] for a in t]
		tk_bad = [a for a in tk if a not in c_kwargs]
		if set(tk_bad) != set(tk[:len(tk_bad)]): # permit non-kw args to contain '='
			die(1, f'{tk_bad[-1]!r}: illegal keyword argument')
		u_kwargs = dict(t[len(tk_bad):])
		u_args = cmd_args[:-len(u_kwargs) or None]
	elif u_nkwargs > 0:
		u_kwargs = dict([a.split('=', 1) for a in cmd_args[len(c_args):] if '=' in a])
		if len(u_kwargs) != u_nkwargs:
			usage_die(f'Command requires exactly {len(c_args)} non-keyword argument{suf(c_args)}')
		if len(u_kwargs) > len(c_kwargs):
			usage_die(f'Command accepts no more than {len(c_kwargs)} keyword argument{suf(c_kwargs)}')

	for k in u_kwargs:
		if k not in c_kwargs:
			usage_die(f'{k!r}: invalid keyword argument')

	def conv_type(arg, arg_name, arg_type):
		if arg_type == 'bytes' and not isinstance(arg, bytes):
			die(1, "'Binary input data must be supplied via STDIN")

		if have_stdin_input and arg_type == 'str' and isinstance(arg, bytes):
			NL = '\r\n' if sys.platform == 'win32' else '\n'
			arg = arg.decode()
			if arg[-len(NL):] == NL: # rstrip one newline
				arg = arg[:-len(NL)]

		if arg_type == 'bool':
			if arg.lower() in ('true', 'yes', '1', 'on'):
				arg = True
			elif arg.lower() in ('false', 'no', '0', 'off'):
				arg = False
			else:
				usage_die(f'{arg!r}: invalid boolean value for keyword argument')

		try:
			return __builtins__[arg_type](arg)
		except:
			die(1, f'{arg!r}: Invalid argument for argument {arg_name} ({arg_type!r} required)')

	if flag == 'VAR_ARGS':
		args = [conv_type(u_args[i], c_args[0][0], c_args[0][1]) for i in range(len(u_args))]
	else:
		args = [conv_type(u_args[i], c_args[i][0], c_args[i][1]) for i in range(len(c_args))]
	kwargs = {k:conv_type(v, k, c_kwargs_types[k].__name__) for k, v in u_kwargs.items()}

	return (args, kwargs)

def process_result(ret, *, pager=False, print_result=False):
	"""
	Convert result to something suitable for output to screen and return it.
	If result is bytes and not convertible to utf8, output as binary using os.write().
	If 'print_result' is True, send the converted result directly to screen or
	pager instead of returning it.
	"""

	def triage_result(o):
		if print_result:
			if pager:
				from .ui import do_pager
				do_pager(o)
			else:
				Msg(o)
		else:
			return o

	if ret is True:
		return True
	elif ret in (False, None):
		die(2, f'tool command returned {ret!r}')
	elif isinstance(ret, str):
		return triage_result(ret)
	elif isinstance(ret, int):
		return triage_result(str(ret))
	elif isinstance(ret, tuple):
		return triage_result('\n'.join([r.decode() if isinstance(r, bytes) else r for r in ret]))
	elif isinstance(ret, bytes):
		try:
			return triage_result(ret.decode())
		except:
			# don't add NL to binary data if it can't be converted to utf8
			return os.write(1, ret) if print_result else ret
	else:
		die(2, f'tool.py: can’t handle return value of type {type(ret).__name__!r}')

def get_cmd_cls(cmd):
	for modname, cmdlist in mods.items():
		if cmd in cmdlist:
			return getattr(importlib.import_module(f'mmgen.tool.{modname}'), 'tool_cmd')
	return False

def get_mod_cls(modname):
	return getattr(importlib.import_module(f'mmgen.tool.{modname}'), 'tool_cmd')

if gc.prog_name.endswith('-tool'):

	cfg = Config(opts_data=opts_data, parse_only=True)
	po = cfg._parsed_opts

	if po.user_opts.get('list'):
		def gen():
			for mod, cmdlist in mods.items():
				if mod == 'help':
					continue
				yield capfirst(get_mod_cls(mod).__doc__.lstrip().split('\n')[0]) + ':'
				for cmd in cmdlist:
					yield '  ' + cmd
				yield ''
		Msg('\n'.join(gen()).rstrip())
		sys.exit(0)

	if len(po.cmd_args) < 1:
		cfg._usage()

	cmd = po.cmd_args[0]

	cls = get_cmd_cls(cmd)

	if not cls:
		die(1, f'{cmd!r}: no such command')

	cfg = Config(
		opts_data   = opts_data,
		parsed_opts = po,
		need_proto  = cls.need_proto,
		init_opts   = {'rpc_backend':'aiohttp'} if cmd == 'twimport' else None,
		process_opts = True)

	cmd, *args = cfg._args

	if cmd in ('help', 'usage') and args:
		args[0] = 'command_name=' + args[0]

	args, kwargs = process_args(cmd, args, cls)

	ret = getattr(cls(cfg, cmdname=cmd), cmd)(*args, **kwargs)

	if type(ret).__name__ == 'coroutine':
		ret = async_run(ret)

	process_result(
		ret,
		pager = kwargs.get('pager'),
		print_result = True)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test/cmdtest.py: Command test runner for the MMGen wallet system
"""

def check_segwit_opts(proto):
	for k, m in (('segwit', 'S'), ('segwit_random', 'S'), ('bech32', 'B')):
		if getattr(cfg, k) and m not in proto.mmtypes:
			die(1, f'--{k.replace("_", "-")} option incompatible with {proto.cls_name}')

def create_shm_dir(data_dir, trash_dir):
	# Laggy flash media can cause pexpect to fail, so create a temporary directory
	# under '/dev/shm' and put datadir and tmpdirs here.
	import shutil
	from subprocess import run
	if sys.platform in ('win32', 'darwin'):
		for tdir in (data_dir, trash_dir):
			try:
				os.listdir(tdir)
			except:
				pass
			else:
				try:
					shutil.rmtree(tdir)
				except: # we couldn't remove data dir - perhaps regtest daemon is running
					try:
						run(['python3', os.path.join('cmds', 'mmgen-regtest'), 'stop'], check=True)
					except:
						die(4, f'Unable to remove {tdir!r}!')
					else:
						time.sleep(2)
						shutil.rmtree(tdir)
			os.mkdir(tdir, 0o755)
		shm_dir = 'test'
	else:
		tdir, pfx = '/dev/shm', 'mmgen-test-'
		try:
			run(f'rm -rf {tdir}/{pfx}*', shell=True, check=True)
		except Exception as e:
			die(2, f'Unable to delete directory tree {tdir}/{pfx}* ({e.args[0]})')
		try:
			import tempfile
			shm_dir = str(tempfile.mkdtemp('', pfx, tdir))
		except Exception as e:
			die(2, f'Unable to create temporary directory in {tdir} ({e.args[0]})')

		dest = os.path.join(shm_dir, os.path.basename(trash_dir))
		os.mkdir(dest, 0o755)

		run(f'rm -rf {trash_dir}', shell=True, check=True)
		os.symlink(dest, trash_dir)

		dest = os.path.join(shm_dir, os.path.basename(data_dir))
		shutil.move(data_dir, dest) # data_dir was created by Config()
		os.symlink(dest, data_dir)

	return shm_dir

import sys, os, time

# overlay must be set up before importing mmgen mods!
try:
	from include.test_init import repo_root
except ImportError:
	from test.include.test_init import repo_root

from mmgen.cfg import Config
from mmgen.color import red, yellow, green, blue, init_color
from mmgen.util import msg, Msg, rmsg, die

from test.include.common import (
	set_globals,
	cmdtest_py_log_fn,
	cmdtest_py_error_fn,
	mk_tmpdir,
	stop_test_daemons)

try:
	os.unlink(os.path.join(repo_root, cmdtest_py_error_fn))
except:
	pass

os.environ['MMGEN_QUIET'] = '0' # for this script and spawned scripts

opts_data = {
	'sets': [
		('list_current_cmd_groups', True, 'list_cmd_groups', True),
		('demo', True, 'exact_output', True),
		('demo', True, 'buf_keypress', True),
		('demo', True, 'pexpect_spawn', True),
	],
	'text': {
		'desc': 'High-level tests for the MMGen Wallet suite',
		'usage':'[options] [command [..command]] | [command_group[.command_subgroup][:command]]',
		'options': """
-h, --help           Print this help message
--, --longhelp       Print help message for long (global) options
-a, --no-altcoin     Skip altcoin tests (WIP)
-A, --no-daemon-autostart Don't start and stop daemons automatically
-B, --bech32         Generate and use Bech32 addresses
-b, --buf-keypress   Use buffered keypresses as with real human input
                     (often required on slow systems, or under emulation)
-c, --print-cmdline  Print the command line of each spawned command
-C, --coverage       Produce code coverage info using trace module
-x, --debug-pexpect  Produce debugging output for pexpect calls
--, --demo           Add extra delay after each send to make input visible.
                     Implies --exact-output --pexpect-spawn --buf-keypress
-d, --deps-only      Run a command or command subgroup’s dependencies without
                     running the command or command group itself.
-D, --no-daemon-stop Don't stop auto-started daemons after running tests
-E, --direct-exec    Bypass pexpect and execute a command directly (for
                     debugging only)
-e, --exact-output   Show the exact output of the MMGen script(s) being run
-G, --exclude-groups=G Exclude the specified command groups (comma-separated)
-k, --devnet-block-period=N Block time for Ethereum devnet bump tests
-l, --list-cmds      List the test script’s available commands
-L, --list-cmd-groups List the test script’s command groups and subgroups
-g, --list-current-cmd-groups List command groups for current configuration
-n, --names          Display command names instead of descriptions
-N, --no-timings     Suppress display of timing information
-o, --log            Log commands to file {lf!r}
-O, --pexpect-spawn  Use pexpect.spawn instead of popen_spawn (much slower,
                     kut does real terminal emulation)
-p, --pause          Pause between tests, resuming on keypress
-P, --profile        Record the execution time of each script
-q, --quiet          Produce minimal output.  Suppress dependency info
-r, --resume=c       Resume at command 'c' after interrupted run
-R, --resume-after=c Same, but resume at command following 'c'
-t, --step           After resuming, execute one command and stop
-S, --skip-deps      Skip dependency checking for command
-u, --usr-random     Get random data interactively from user
-T, --pexpect-timeout=T Set the timeout for pexpect
-v, --verbose        Produce more verbose output
-W, --no-dw-delete   Don't remove default wallet from data dir after dw tests
                     are done
-X, --exit-after=C   Exit after command 'C'
-y, --segwit         Generate and use Segwit addresses
-Y, --segwit-random  Generate and use a random mix of Segwit and Legacy addrs
""",
		'notes': """

If no command is given, the whole test suite is run for the currently
specified coin (default BTC).

For traceback output and error file support, set the EXEC_WRAPPER_TRACEBACK
environment var
"""
	},
	'code': {
		'options': lambda proto, help_notes, s: s.format(
				lf = cmdtest_py_log_fn
			)
	}
}

# we need some opt values before running opts.init, so parse without initializing:
po = Config(opts_data=opts_data, parse_only=True)._parsed_opts

data_dir = Config.test_datadir

# step 1: delete data_dir symlink in ./test;
if not po.user_opts.get('skip_deps'):
	try:
		os.unlink(data_dir)
	except:
		pass

# step 2: opts.init will create new data_dir in ./test (if not po.user_opts['skip_deps'])
cfg = Config(opts_data=opts_data)

if cfg.no_altcoin and cfg.coin != 'BTC':
	die(1, f'--no-altcoin incompatible with --coin={cfg.coin}')

set_globals(cfg)

type(cfg)._reset_ok += (
	'no_daemon_autostart',
	'names',
	'no_timings',
	'exit_after',
	'resuming',
	'skipping_deps')

cfg.resuming = any(k in po.user_opts for k in ('resume', 'resume_after'))
cfg.skipping_deps = cfg.resuming or 'skip_deps' in po.user_opts

cmd_args = cfg._args

if cfg.pexpect_spawn and sys.platform == 'win32':
	die(1, '--pexpect-spawn option not supported on Windows platform, exiting')

if cfg.daemon_id and cfg.daemon_id in cfg.blacklisted_daemons.split():
	die(1, f'cmdtest.py: daemon {cfg.daemon_id!r} blacklisted, exiting')

# step 3: move data_dir to /dev/shm and symlink it back to ./test:
trash_dir = os.path.join('test', 'trash')
trash_dir2 = os.path.join('test', 'trash2')

if not cfg.skipping_deps:
	shm_dir = create_shm_dir(data_dir, trash_dir)

check_segwit_opts(cfg._proto)

testing_segwit = cfg.segwit or cfg.segwit_random or cfg.bech32

if cfg.test_suite_deterministic:
	cfg.no_timings = True
	init_color(num_colors=0)
	os.environ['MMGEN_DISABLE_COLOR'] = '1' # for this script and spawned scripts

if cfg.profile:
	cfg.names = True

if cfg.exact_output:
	qmsg = qmsg_r = lambda s: None
else:
	qmsg = cfg._util.qmsg
	qmsg_r = cfg._util.qmsg_r

if cfg.skipping_deps:
	cfg.no_daemon_autostart = True

from test.cmdtest_d.include.cfg import cfgs

def list_cmds():

	def gen_output():

		from test.cmdtest_d.include.group_mgr import CmdGroupMgr
		gm = CmdGroupMgr(cfg)
		cw, d = 0, []

		yield green('AVAILABLE COMMANDS:')

		for gname in gm.cmd_groups:
			tg = gm.gm_init_group(cfg, None, gname, None, None)
			gdesc = tg.__doc__.strip() if tg.__doc__ else type(tg).__name__
			d.append((gname, gdesc, gm.cmd_list, gm.dpy_data))
			cw = max(max(len(k) for k in gm.dpy_data), cw)

		for gname, gdesc, cmd_list, dpy_data in d:
			yield '\n'+green(f'{gname!r} - {gdesc}:')
			for cmd in cmd_list:
				data = dpy_data[cmd]
				yield '    {:{w}} - {}'.format(
					cmd,
					(data if isinstance(data, str) else data[1]),
					w = cw)

	from mmgen.ui import do_pager
	do_pager('\n'.join(gen_output()))

def create_tmp_dirs(shm_dir):
	if sys.platform in ('win32', 'darwin'):
		for cfg in sorted(cfgs):
			mk_tmpdir(cfgs[cfg]['tmpdir'])
	else:
		os.makedirs(os.path.join('test', 'tmp'), mode=0o755, exist_ok=True)
		for cfg in sorted(cfgs):
			src = os.path.join(shm_dir, cfgs[cfg]['tmpdir'].split('/')[-1])
			mk_tmpdir(src)
			try:
				os.unlink(cfgs[cfg]['tmpdir'])
			except OSError as e:
				if e.errno != 2:
					raise
			finally:
				os.symlink(src, cfgs[cfg]['tmpdir'])

def set_restore_term_at_exit():
	import termios, atexit
	fd = sys.stdin.fileno()
	old = termios.tcgetattr(fd)
	def at_exit():
		termios.tcsetattr(fd, termios.TCSADRAIN, old)
	atexit.register(at_exit)

if __name__ == '__main__':

	if not cfg.skipping_deps: # do this before list cmds exit, so we stay in sync with shm_dir
		create_tmp_dirs(shm_dir)

	if cfg.list_cmd_groups:
		from test.cmdtest_d.include.group_mgr import CmdGroupMgr
		CmdGroupMgr(cfg).list_cmd_groups()
		sys.exit(0)
	elif cfg.list_cmds:
		list_cmds()
		sys.exit(0)

	if cfg.pause:
		set_restore_term_at_exit()

	from mmgen.exception import TestSuiteSpawnedScriptException
	from test.cmdtest_d.include.runner import CmdTestRunner

	try:
		tr = CmdTestRunner(cfg, repo_root, data_dir, trash_dir, trash_dir2)
		tr.run_tests(cmd_args)
		tr.print_warnings()
		if tr.daemon_started and not cfg.no_daemon_stop:
			stop_test_daemons(tr.network_id, remove_datadir=True)
		if hasattr(tr, 'tg'):
			del tr.tg
		del tr
	except KeyboardInterrupt:
		if tr.daemon_started and not cfg.no_daemon_stop:
			stop_test_daemons(tr.network_id, remove_datadir=True)
		tr.print_warnings()
		if hasattr(tr, 'tg'):
			del tr.tg
		del tr
		die(1, yellow('\ntest.py exiting at user request'))
	except TestSuiteSpawnedScriptException as e:
		# if spawned script is not running under exec_wrapper, output brief error msg:
		if os.getenv('MMGEN_EXEC_WRAPPER'):
			Msg(red(str(e)))
			Msg(blue('cmdtest.py: spawned script exited with error'))
		if hasattr(tr, 'tg'):
			del tr.tg
		del tr
		raise
	except Exception as e:
		if type(e).__name__ == 'TestSuiteException':
			rmsg('TEST ERROR: ' + str(e))
		if hasattr(tr, 'tg'):
			del tr.tg
		del tr
		# if cmdtest.py itself is running under exec_wrapper, re-raise so exec_wrapper can handle exception:
		if os.getenv('MMGEN_EXEC_WRAPPER') or not os.getenv('MMGEN_IGNORE_TEST_PY_EXCEPTION'):
			raise
		die(1, red('Test script exited with error'))



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tw.addresses: Bitcoin base protocol tracking wallet address list class
"""

from ....tw.addresses import TwAddresses
from ....tw.shared import TwLabel
from ....obj import get_obj

from .rpc import BitcoinTwRPC
from .view import BitcoinTwView

class BitcoinTwAddresses(BitcoinTwView, TwAddresses, BitcoinTwRPC):

	has_age = True
	has_used = True

	prompt_fs_in = [
		'Sort options: [a]mt, [A]ge, [M]mgen addr, [r]everse',
		'Column options: toggle [D]ays/date/confs/block',
		'Filters: show [E]mpty addrs, [u]sed addrs, all [L]abels',
		'View/Print: pager [v]iew, [w]ide pager view, [p]rint{s}',
		'Actions: [q]uit menu, r[e]draw, add [l]abel:']
	prompt_fs_repl = {
		'BCH': (1, 'Column options: toggle [D]ays/date/confs/block, cas[h]addr')
	}
	extra_key_mappings = {
		'A':'s_age',
		'D':'d_days',
		'u':'d_showused'}

	async def get_rpc_data(self):

		qmsg = self.cfg._util.qmsg
		qmsg_r = self.cfg._util.qmsg_r
		qmsg_r('Getting unspent outputs...')
		addrs = await self.get_unspent_by_mmid(minconf=self.minconf)
		qmsg('done')

		coin_amt = self.proto.coin_amt
		amt0 = coin_amt('0')
		self.total = sum((v['amt'] for v in addrs.values()), start=amt0)

		qmsg_r('Getting labels and associated addresses...')
		for e in await self.get_label_addr_pairs():
			if e.label and e.label.mmid not in addrs:
				addrs[e.label.mmid] = {
					'addr':   e.coinaddr,
					'amt':    amt0,
					'recvd':  amt0,
					'is_used': False,
					'confs':  0,
					'lbl':    e.label}
		qmsg('done')

		qmsg_r('Getting received funds data...')
		# args: 1:minconf, 2:include_empty, 3:include_watchonly, 4:include_immature_coinbase (>=v23.0.0)
		for d in await self.rpc.call('listreceivedbylabel', 1, True, True):
			label = get_obj(TwLabel, proto=self.proto, text=d['label'])
			if label:
				assert label.mmid in addrs, f'{label.mmid!r} not found in addrlist!'
				amt = coin_amt(d['amount'])
				addrs[label.mmid]['recvd'] = amt
				addrs[label.mmid]['is_used'] = bool(amt)
				addrs[label.mmid]['confs'] = d['confirmations']
		qmsg('done')

		return addrs



#!/usr/bin/env python3

import sys, os
os.chdir(os.path.dirname(os.path.dirname(os.path.abspath(os.path.dirname(sys.argv[0])))))
sys.path[0] = os.curdir

from mmgen.cfg import Config
cfg = Config()

from mmgen.util import msg
from mmgen.term import init_term, get_term
init_term(cfg)
term = get_term()

if sys.argv[1] == 'echo':

	from mmgen.ui import line_input
	from mmgen.term import get_char_raw

	def test_noecho():
		term.init(noecho=True)
		ret = line_input(cfg, 'noecho> ')
		msg(f'==> [{ret.upper()}]')
		get_char_raw()

	def test_echo():
		term.set('echo')
		ret = line_input(cfg, 'echo> ')
		msg(f'==> [{ret.upper()}]')

	test_noecho()
	test_echo()
	test_noecho()

elif sys.argv[1] == 'cleanup':

	term.register_cleanup()

	import tty
	tty.setcbreak(term.stdin_fd)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
crypto: Random number, password hashing and symmetric encryption routines for the MMGen suite
"""

import os
from collections import namedtuple

from .cfg import gc
from .util import msg, msg_r, ymsg, fmt, die, make_chksum_8, oneshot_warning

class Crypto:

	mmenc_ext = 'mmenc'
	scramble_hash_rounds = 10

	salt_len       = 16
	aesctr_iv_len  = 16
	aesctr_dfl_iv  = int.to_bytes(1, aesctr_iv_len, 'big')
	hincog_chk_len = 8

	mmenc_salt_len = 32
	mmenc_nonce_len = 32

	# Scrypt params: 'id_num': [N, r, p] (N is an exponent of two)
	# NB: hashlib.scrypt in Python (>=v3.6) supports max N value of 14.  This means that
	# for hash presets > 3 the standalone scrypt library must be used!
	_hp = namedtuple('scrypt_preset', ['N', 'r', 'p'])
	hash_presets = {
		'1': _hp(12, 8, 1),
		'2': _hp(13, 8, 4),
		'3': _hp(14, 8, 8),
		'4': _hp(15, 8, 12),
		'5': _hp(16, 8, 16),
		'6': _hp(17, 8, 20),
		'7': _hp(18, 8, 24),
	}

	class pwfile_reuse_warning(oneshot_warning):
		message = 'Reusing passphrase from file {!r} at user request'
		def __init__(self, fn):
			oneshot_warning.__init__(self, div=fn, fmt_args=[fn], reverse=True)

	def pwfile_used(self, passwd_file):
		if hasattr(self, '_pwfile_used'):
			self.pwfile_reuse_warning(passwd_file)
			return True
		else:
			self._pwfile_used = True
			return False

	def __init__(self, cfg):
		self.cfg = cfg
		self.util = cfg._util

	def get_hash_params(self, hash_preset):
		if hash_preset in self.hash_presets:
			return self.hash_presets[hash_preset] # N, r, p
		else: # Shouldn't be here
			die(3, f"{hash_preset}: invalid 'hash_preset' value")

	def sha256_rounds(self, s):
		from hashlib import sha256
		for _ in range(self.scramble_hash_rounds):
			s = sha256(s).digest()
		return s

	def scramble_seed(self, seed, scramble_key):
		import hmac
		step1 = hmac.digest(seed, scramble_key, 'sha256')
		if self.cfg.debug:
			msg(f'Seed:  {seed.hex()!r}\nScramble key: {scramble_key}\nScrambled seed: {step1.hex()}\n')
		return self.sha256_rounds(step1)

	def encrypt_seed(self, data, key, *, desc='seed'):
		return self.encrypt_data(data, key=key, desc=desc)

	def decrypt_seed(self, enc_seed, key, *, seed_id, key_id):
		self.util.vmsg_r('Checking key...')
		chk1 = make_chksum_8(key)
		if key_id:
			if not self.util.compare_chksums(key_id, 'key ID', chk1, 'computed'):
				msg('Incorrect passphrase or hash preset')
				return False

		dec_seed = self.decrypt_data(enc_seed, key, desc='seed')
		chk2     = make_chksum_8(dec_seed)
		if seed_id:
			if self.util.compare_chksums(seed_id, 'Seed ID', chk2, 'decrypted seed'):
				self.util.qmsg('Passphrase is OK')
			else:
				if not self.cfg.debug:
					msg_r('Checking key ID...')
					if self.util.compare_chksums(key_id, 'key ID', chk1, 'computed'):
						msg('Key ID is correct but decryption of seed failed')
					else:
						msg('Incorrect passphrase or hash preset')
				self.util.vmsg('')
				return False

		self.util.dmsg(f'Decrypted seed: {dec_seed.hex()}')
		return dec_seed

	def encrypt_data(
			self,
			data,
			*,
			key,
			iv     = aesctr_dfl_iv,
			desc   = 'data',
			verify = True,
			silent = False):

		from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
		from cryptography.hazmat.backends import default_backend
		if not silent:
			self.util.vmsg(f'Encrypting {desc}')
		c = Cipher(algorithms.AES(key), modes.CTR(iv), backend=default_backend())
		encryptor = c.encryptor()
		enc_data = encryptor.update(data) + encryptor.finalize()

		if verify:
			self.util.vmsg_r(f'Performing a test decryption of the {desc}...')
			c = Cipher(algorithms.AES(key), modes.CTR(iv), backend=default_backend())
			encryptor = c.encryptor()
			dec_data = encryptor.update(enc_data) + encryptor.finalize()
			if dec_data != data:
				die(2, f'ERROR.\nDecrypted {desc} doesn’t match original {desc}')
			if not silent:
				self.util.vmsg('done')

		return enc_data

	def decrypt_data(
			self,
			enc_data,
			key,
			*,
			iv   = aesctr_dfl_iv,
			desc = 'data'):

		from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
		from cryptography.hazmat.backends import default_backend
		self.util.vmsg_r(f'Decrypting {desc} with key...')
		c = Cipher(algorithms.AES(key), modes.CTR(iv), backend=default_backend())
		encryptor = c.encryptor()
		return encryptor.update(enc_data) + encryptor.finalize()

	def scrypt_hash_passphrase(
			self,
			passwd,
			salt,
			hash_preset,
			*,
			buflen = 32):

		# Buflen arg is for brainwallets only, which use this function to generate
		# the seed directly.
		ps = self.get_hash_params(hash_preset)

		if isinstance(passwd, str):
			passwd = passwd.encode()

		def do_hashlib_scrypt():
			from hashlib import scrypt
			return scrypt(
				password = passwd,
				salt     = salt,
				n        = 2**ps.N,
				r        = ps.r,
				p        = ps.p,
				maxmem   = 0,
				dklen    = buflen)

		def do_standalone_scrypt():
			import scrypt
			return scrypt.hash(
				password = passwd,
				salt     = salt,
				N        = 2**ps.N,
				r        = ps.r,
				p        = ps.p,
				buflen   = buflen)

		if int(hash_preset) > 3:
			msg_r('Hashing passphrase, please wait...')

		# hashlib.scrypt doesn't support N > 14 (hash preset > 3)
		ret = (
			do_standalone_scrypt() if ps.N > 14 or self.cfg.force_standalone_scrypt_module else
			do_hashlib_scrypt())

		if int(hash_preset) > 3:
			msg_r('\b'*34 + ' '*34 + '\b'*34)

		return ret

	def make_key(
			self,
			passwd,
			salt,
			hash_preset,
			*,
			desc      = 'encryption key',
			from_what = 'passphrase',
			verbose   = False):

		if self.cfg.verbose or verbose:
			msg_r(f"Generating {desc}{' from ' + from_what if from_what else ''}...")
		key = self.scrypt_hash_passphrase(passwd, salt, hash_preset)
		if self.cfg.verbose or verbose:
			msg('done')
		self.util.dmsg(f'Key: {key.hex()}')
		return key

	def _get_random_data_from_user(self, uchars=None, *, desc='data'):

		if uchars is None:
			uchars = self.cfg.usr_randchars

		info1 = f"""
			Now we're going to gather some additional input from the keyboard to further
			randomize the random data {desc}.

			An encryption key will be created from this input, and the random data will
			be encrypted using the key.  The resulting data is guaranteed to be at least
			as random as the original random data, so even if you type very predictably
			no harm will be done.

			However, to gain the maximum benefit, try making your input as random as
			possible.  Type slowly and choose your symbols carefully.  Try to use both
			upper and lowercase letters as well as punctuation and numerals. The timings
			between your keystrokes will also be used as a source of entropy, so be as
			random as possible in your timing as well.
		"""
		info2 = f"""
			Please type {uchars} symbols on your keyboard.  What you type will not be displayed
			on the screen.
		"""

		msg(f'Enter {uchars} random symbols' if self.cfg.quiet else
			'\n' + fmt(info1, indent='  ') +
			'\n' + fmt(info2))

		import time
		from .term import get_char_raw
		key_data = ''
		time_data = []

		for i in range(uchars):
			key_data += get_char_raw(f'\rYou may begin typing.  {uchars-i} symbols left: ')
			time_data.append(time.time())

		msg_r('\r' if self.cfg.quiet else f'\rThank you.  That’s enough.{" "*18}\n\n')

		time_data = [f'{t:.22f}'.rstrip('0') for t in time_data]

		avg_prec = sum(len(t.split('.')[1]) for t in time_data) // len(time_data)

		if avg_prec < gc.min_time_precision:
			ymsg(f'WARNING: Avg. time precision of only {avg_prec} decimal points. User entropy quality is degraded!')

		ret = key_data + '\n' + '\n'.join(time_data)

		if self.cfg.debug:
			msg(f'USER ENTROPY (user input + keystroke timings):\n{ret}')

		from .ui import line_input
		line_input(self.cfg, 'User random data successfully acquired.  Press ENTER to continue: ')

		return ret.encode()

	def get_random(self, length):

		os_rand = os.urandom(length)
		assert len(os_rand) == length, f'OS random number generator returned {len(os_rand)} (!= {length}) bytes!'

		return self.add_user_random(
			rand_bytes = os_rand,
			desc       = 'from your operating system')

	def add_user_random(
			self,
			rand_bytes,
			*,
			desc,
			urand = {'data':b'', 'counter':0}):

		assert type(rand_bytes) is bytes, 'add_user_random_chk1'

		if self.cfg.usr_randchars:

			if not urand['data']:
				from hashlib import sha256
				urand['data'] = sha256(self._get_random_data_from_user(desc=desc)).digest()

			# counter protects against very evil rng that might repeatedly output the same data
			urand['counter'] += 1

			os_rand = os.urandom(8)
			assert len(os_rand) == 8, f'OS random number generator returned {len(os_rand)} (!= 8) bytes!'

			import hmac
			key = hmac.digest(
				urand['data'],
				os_rand + int.to_bytes(urand['counter'], 8, 'big'),
				'sha256')

			msg(f'Encrypting random data {desc} with ephemeral key #{urand["counter"]}')

			return self.encrypt_data(data=rand_bytes, key=key, desc=desc, verify=False, silent=True)
		else:
			return rand_bytes

	def get_hash_preset_from_user(
			self,
			old_preset = gc.dfl_hash_preset,
			*,
			data_desc  = 'data',
			prompt     = None):

		prompt = prompt or (
			f'Enter hash preset for {data_desc}, \n' +
			f'or hit ENTER to accept the default value ({old_preset!r}): ')

		from .ui import line_input
		while True:
			ret = line_input(self.cfg, prompt)
			if ret:
				if ret in self.hash_presets:
					return ret
				else:
					msg('Invalid input.  Valid choices are {}'.format(', '.join(self.hash_presets)))
			else:
				return old_preset

	def get_new_passphrase(self, data_desc, hash_preset, passwd_file, *, pw_desc='passphrase'):
		message = f"""
				You must choose a passphrase to encrypt your {data_desc} with.
				A key will be generated from your passphrase using a hash preset of '{hash_preset}'.
				Please note that no strength checking of passphrases is performed.
				For an empty passphrase, just hit ENTER twice.
			"""
		if passwd_file:
			from .fileutil import get_words_from_file
			pw = ' '.join(get_words_from_file(
				cfg    = self.cfg,
				infile = passwd_file,
				desc   = f'{pw_desc} for {data_desc}',
				quiet  = self.pwfile_used(passwd_file)))
		else:
			self.util.qmsg('\n'+fmt(message, indent='  '))
			from .ui import get_words_from_user
			if self.cfg.echo_passphrase:
				pw = ' '.join(get_words_from_user(self.cfg, f'Enter {pw_desc} for {data_desc}: '))
			else:
				for _ in range(gc.passwd_max_tries):
					pw = ' '.join(get_words_from_user(self.cfg, f'Enter {pw_desc} for {data_desc}: '))
					pw_chk = ' '.join(get_words_from_user(self.cfg, f'Repeat {pw_desc}: '))
					self.util.dmsg(f'Passphrases: [{pw}] [{pw_chk}]')
					if pw == pw_chk:
						self.util.vmsg('Passphrases match')
						break
					msg('Passphrases do not match.  Try again.')
				else:
					die(2, f'User failed to duplicate passphrase in {gc.passwd_max_tries} attempts')

		if pw == '':
			self.util.qmsg('WARNING: Empty passphrase')

		return pw

	def get_passphrase(self, data_desc, passwd_file, *, pw_desc='passphrase'):
		if passwd_file:
			from .fileutil import get_words_from_file
			return ' '.join(get_words_from_file(
				cfg    = self.cfg,
				infile = passwd_file,
				desc   = f'{pw_desc} for {data_desc}',
				quiet  = self.pwfile_used(passwd_file)))
		else:
			from .ui import get_words_from_user
			return ' '.join(get_words_from_user(self.cfg, f'Enter {pw_desc} for {data_desc}: '))

	def mmgen_encrypt(self, data, *, passwd=None, desc='data', hash_preset=None):
		salt  = self.get_random(self.mmenc_salt_len)
		iv    = self.get_random(self.aesctr_iv_len)
		nonce = self.get_random(self.mmenc_nonce_len)
		hp    = hash_preset or self.cfg.hash_preset or self.get_hash_preset_from_user(data_desc=desc)
		m     = ('user-requested', 'default')[hp=='3']
		self.util.vmsg(f'Encrypting {desc}')
		self.util.qmsg(f'Using {m} hash preset of {hp!r}')
		passwd = passwd or self.get_new_passphrase(
			data_desc = desc,
			hash_preset = hp,
			passwd_file = self.cfg.passwd_file)
		key    = self.make_key(passwd, salt, hp)
		from hashlib import sha256
		enc_d  = self.encrypt_data(sha256(nonce+data).digest() + nonce + data, key=key, iv=iv, desc=desc)
		return salt+iv+enc_d

	def mmgen_decrypt(self, data, *, passwd=None, desc='data', hash_preset=None):
		self.util.vmsg(f'Preparing to decrypt {desc}')
		dstart = self.mmenc_salt_len + self.aesctr_iv_len
		salt   = data[:self.mmenc_salt_len]
		iv     = data[self.mmenc_salt_len:dstart]
		enc_d  = data[dstart:]
		hp     = hash_preset or self.cfg.hash_preset or self.get_hash_preset_from_user(data_desc=desc)
		m  = ('user-requested', 'default')[hp=='3']
		self.util.qmsg(f'Using {m} hash preset of {hp!r}')
		passwd = passwd or self.get_passphrase(
			data_desc = desc,
			passwd_file = self.cfg.passwd_file)
		key    = self.make_key(passwd, salt, hp)
		dec_d  = self.decrypt_data(enc_d, key, iv=iv, desc=desc)
		sha256_len = 32
		from hashlib import sha256
		if dec_d[:sha256_len] == sha256(dec_d[sha256_len:]).digest():
			self.util.vmsg('OK')
			return dec_d[sha256_len+self.mmenc_nonce_len:]
		else:
			msg('Incorrect passphrase or hash preset')
			return False

	def mmgen_decrypt_retry(self, d, *, desc='data'):
		while True:
			d_dec = self.mmgen_decrypt(d, desc=desc)
			if d_dec:
				return d_dec
			msg('Trying again...')



#!/usr/bin/env python3

import json

from bip_utils.slip.slip44 import Slip44
from bip_utils.bip.conf.bip44.bip44_conf import Bip44Conf
from bip_utils.bip.conf.bip49.bip49_conf import Bip49Conf
from bip_utils.bip.conf.bip84.bip84_conf import Bip84Conf
from bip_utils.bip.conf.bip86.bip86_conf import Bip86Conf

import script_init
from mmgen.cfg import Config
from mmgen.main import launch

opts_data = {
	'text': {
		'desc': 'Aggregate bip_lib and SLIP-44 data into a bip_hd chainparams table',
		'usage':'[opts] infile',
		'options': """
-h, --help               Print this help message.
""",
	'notes': """
source: https://github.com/MetaMask/slip44/blob/main/slip44.json
"""
	}
}

cfg = Config(opts_data=opts_data)

def curve_clsname_abbr(s):
	return {
		'Bip32Slip10Secp256k1':      'x',
		'Bip32Slip10Ed25519':        'edw',
		'Bip32Slip10Ed25519Blake2b': 'blk',
		'Bip32Slip10Nist256p1':      'nist',
		'Bip32KholawEd25519':        'khol',
	}.get(s, s)

fs2 = '{:5} {:2} {:16} {:8} {:8} {:6} {:8} {:8}'
hdr2 = fs2.format('CURVE', 'NW', 'ADDR_CLS', 'VB_PRV', 'VB_PUB', 'VB_WIF', 'VB_ADDR', 'DFL_PATH')

dfl_vb_prv = '0488ade4'
dfl_vb_pub = '0488b21e'
dfl_curve = 'secp'
dfl_dfl_path = "0'/0/0"

def get_bip_utils_data(bipnum, n):
	name, v = bip_utils_data[bipnum][n]
	vb_prv = v.m_key_net_ver.m_priv_net_ver.hex()
	vb_pub = v.m_key_net_ver.m_pub_net_ver.hex()
	ap = v.m_addr_params
	return fs2.format(
		curve_clsname_abbr(v.m_bip32_cls.__name__),
		'T' if v.m_is_testnet else 'm',
		v.m_addr_cls.__name__.removesuffix('AddrEncoder'),
		'x' if vb_prv == dfl_vb_prv else vb_prv,
		'x' if vb_pub == dfl_vb_pub else vb_pub,
		v.m_wif_net_ver.hex() if isinstance(v.m_wif_net_ver, bytes) else '-',
		ap['net_ver'].hex() if 'net_ver' in ap else 'h:'+ap['hrp'] if 'hrp' in ap else 'spec' if ap else '-',
		'x' if v.m_def_path == dfl_dfl_path else v.m_def_path,
	)

def gen():

	def format_data(bipnum, n, sym, name):
		return fs.format(
			n,
			sym if sym else '---',
			get_bip_utils_data(bipnum, n) if bipnum else '-',
			name if name else '---')

	fs = '{:<6} {:6} {:1} {}'

	yield f'[defaults]'
	yield fs.format('IDX', 'CHAIN', hdr2, 'NAME')
	yield fs.format('0', '-', fs2.format(dfl_curve, '-', '-', dfl_vb_prv, dfl_vb_pub, '-', '-', dfl_dfl_path), '-')

	yield f'\n[bip-44]'
	yield fs.format('IDX', 'CHAIN', hdr2, 'NAME')
	for k, v in slip44_data.items():
		if int(k) in bip_utils_data[44]:
			yield format_data(44, int(k), v['symbol'], v['name'])

	for bipnum in (49, 84, 86):
		yield f'\n[bip-{bipnum}]'
		yield fs.format('IDX', 'CHAIN', hdr2, 'NAME')
		for n, v in sorted(bip_utils_data[bipnum].items()):
			nd = v[1].m_coin_names
			yield format_data(bipnum, n, nd.m_abbr, nd.m_name)

	yield f'\n[bip-44-unsupported]'
	yield fs.format('IDX', 'CHAIN', '', 'NAME')
	for k, v in slip44_data.items():
		if not int(k) in bip_utils_data[44]:
			yield format_data(None, int(k), v['symbol'], v['name'])

def main():

	global slip44_data, bip_utils_data

	if len(cfg._args) != 1:
		cfg._usage()

	with open(cfg._args[0]) as fh:
		slip44_data = json.loads(fh.read())

	bip_utils_data = {
		n:{v.m_coin_idx:(k, v) for k, v in globals()[f'Bip{n}Conf'].__dict__.items() if not k.startswith('_')}
			for n in (44, 49, 84, 86)
	}

	print('\n'.join(gen()))

launch(func=main)



#!/usr/bin/env python3

import sys, os

import script_init
from mmgen.main import launch
from mmgen.cfg import Config
from mmgen.util import msg, Msg, make_chksum_6
from mmgen.fileutil import get_lines_from_file

opts_data = {
	'text': {
		'desc': 'Compute checksum for a MMGen data file',
		'usage':'[opts] infile',
		'options': """
-h, --help               Print this help message.
-i, --include-first-line Include the first line of the file (you probably don't want this)
"""
	}
}

cfg = Config(opts_data=opts_data)

def main():
	lines = get_lines_from_file(cfg, cfg._args[0])
	start = (1, 0)[bool(cfg.include_first_line)]
	a = make_chksum_6(' '.join(lines[start:]).encode())
	if start == 1:
		b = lines[0]
		msg(
			'Checksum in file OK' if a == b else
			f"Checksum in file ({b}) doesn't match computed value!")
	Msg(a)

launch(func=main)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tx.completed: completed transaction class
"""

from .base import Base

class Completed(Base):
	"""
	signed or unsigned transaction with associated file
	"""
	filename_api = True

	def __init__(self, cfg, *args, filename=None, data=None, quiet_open=False, **kwargs):

		assert (filename or data) and not (filename and data), 'CompletedTX_chk1'

		super().__init__(cfg=cfg, *args, **kwargs)

		if data:
			self.__dict__ = data | {'twctl': self.twctl}
			self.name = type(self).__name__
		else:
			from .file import MMGenTxFile
			try:
				MMGenTxFile(self).parse(str(filename), quiet_open=quiet_open)
				self.check_serialized_integrity()
			except Exception as e:
				from ..color import orange
				from ..util import msg
				msg(orange(
					f'Something is wrong with transaction file ‘{filename}’\n'
					'To fix this problem, please move or delete the file'))
				raise e

			# repeat with sign and send, because coin daemon could be restarted
			self.check_correct_chain()

			if self.check_sigs() != self.signed:
				from ..util import die
				die(1, 'Transaction is {}signed!'.format('not ' if self.signed else ''))

			self.infile = filename

	@property
	def file(self):
		from .file import MMGenTxFile
		return MMGenTxFile(self)

	@staticmethod
	def ext_to_cls(ext, proto):
		"""
		see twctl:import_token()
		"""
		from .unsigned import Unsigned, AutomountUnsigned
		from .online import Sent, AutomountSent
		for cls in (Unsigned, AutomountUnsigned, Sent, AutomountSent):
			if ext == getattr(cls, 'ext'):
				return cls

		if proto.tokensym:
			from .online import OnlineSigned as Signed
			from .online import AutomountOnlineSigned as AutomountSigned
		else:
			from .signed import Signed, AutomountSigned
		for cls in (Signed, AutomountSigned):
			if ext == getattr(cls, 'ext'):
				return cls

	def check_swap_memo(self):
		if memo_bytes := self.get_swap_memo_maybe():
			from ..swap.proto.thorchain import Memo
			if Memo.is_partial_memo(memo_bytes):
				from ..protocol import init_proto
				text = memo_bytes.decode('ascii')
				p = Memo.parse(text)
				r = self.recv_asset
				assert p.function == 'SWAP', f'‘{p.function}’: unsupported function in swap memo ‘{text}’'
				assert p.asset.name == r.name, f'invalid memo: {p.asset.name} != {r.name}'
				proto = init_proto(
						self.cfg,
						r.coin,
						network = self.cfg.network,
						tokensym = r.tokensym,
						need_amt = True)
				if mmid := getattr(self, 'swap_recv_addr_mmid', None):
					pass
				elif self.cfg.allow_non_wallet_swap:
					from ..util import ymsg
					ymsg('Warning: allowing swap to non-wallet address (--allow-non-wallet-swap)')
				else:
					raise ValueError('Swap to non-wallet address forbidden (override with --allow-non-wallet-swap)')
				return self.Output(proto, addr=p.address, mmid=mmid, amt=proto.coin_amt('0'))

		if self.is_swap:
			raise ValueError('missing or invalid memo in swap transaction')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
altcoin.py - Constants for Bitcoin-derived altcoins
"""

# Sources:
#   lb: https://github.com/libbitcoin/libbitcoin/wiki/Altcoin-Version-Mappings
#   pc: https://github.com/richardkiss/pycoin/blob/master/pycoin/networks/legacy_networks.py
#   vg: https://github.com/exploitagency/vanitygen-plus/blob/master/keyconv.c
#   wn: https://walletgenerator.net
#   cc: https://www.cryptocompare.com/api/data/coinlist/ (names, symbols only)

# BIP44
# https://github.com/bitcoin/bips/blob/master/bip-0044.mediawiki
# BIP44 registered coin types: https://github.com/satoshilabs/slips/blob/master/slip-0044.md

# WIP:
#   NSR:  149/191 c/u,  63/('S'),  64/('S'|'T')
#   NBT:  150/191 c/u,  25/('B'),  26/('B')

from collections import namedtuple

from ..cfg import gc
from ..protocol import CoinProtocol
from ..proto.btc.params import mainnet

ce = namedtuple('CoinInfoEntry',
	['name', 'symbol', 'wif_ver_num', 'p2pkh_info', 'p2sh_info', 'has_segwit', 'trust_level'])

class CoinInfo:
	coin_constants = {}
	coin_constants['mainnet'] = (
#   Trust levels: -1=disabled 0=untested 1=low 2=med 3=high 4=very high (no warn) 5=unconditional
#   Trust levels apply to key/address generation only.
#   Non core-coin fork coins (i.e. BCG) must be disabled here to prevent generation from
#   incorrect scrambled seed.
	ce('Bitcoin',               'BTC',     0x80,   (0x00,'1'),       (0x05,'3'),       True,  5),
	ce('BitcoinCashNode',       'BCH',     0x80,   (0x00,'1'),       (0x05,'3'),       False, 5),
	ce('BitcoinGold',           'BCG',     0x80,   (0x00,'1'),       (0x05,'3'),       True, -1),
	ce('2GiveCoin',             '2GIVE',   0xa7,   (0x27,('G','H')), None,             False, 0),
	ce('42Coin',                '42',      0x88,   (0x08,'4'),       None,             False, 1),
	ce('ACoin',                 'ACOIN',   0xe6,   (0x17,'A'),       None,             False, 0),
	ce('Alphacoin',             'ALF',     0xd2,   (0x52,('Z','a')), None,             False, 0),
	ce('Anoncoin',              'ANC',     0x97,   (0x17,'A'),       None,             False, 1),
	ce('Apexcoin',              'APEX',    0x97,   (0x17,'A'),       None,             False, 0),
	ce('Aquariuscoin',          'ARCO',    0x97,   (0x17,'A'),       None,             False, 0),
	ce('Argentum',              'ARG',     0x97,   (0x17,'A'),       (0x05,'3'),       False, 1),
	ce('AsiaCoin',              'AC',      0x97,   (0x17,'A'),       (0x08,'4'),       False, 1),
	ce('Auroracoin',            'AUR',     0x97,   (0x17,'A'),       None,             False, 1),
	ce('BBQcoin',               'BQC',     0xd5,   (0x55,'b'),       None,             False, 1),
	ce('BitcoinDark',           'BTCD',    0xbc,   (0x3c,'R'),       (0x55,'b'),       False, 1),
	ce('BitcoinFast',           'BCF',     0xe0,   (0x60,('f','g')), None,             False, 0),
	ce('BitQuark',              'BTQ',     0xba,   (0x3a,'Q'),       None,             False, 0),
	ce('Blackcoin',             'BLK',     0x99,   (0x19,'B'),       (0x55,'b'),       False, 1),
	ce('BlackmoonCrypto',       'BMC',     0x83,   (0x03,'2'),       None,             False, 0),
	ce('BlockCat',              'CAT',     0x95,   (0x15,'9'),       None,             False, 0),
	ce('CanadaECoin',           'CDN',     0x9c,   (0x1c,'C'),       (0x05,'3'),       False, 1),
	ce('CannabisCoin',          'CANN',    0x9c,   (0x1c,'C'),       None,             False, 0),
	ce('CannaCoin',             'CCN',     0x9c,   (0x1c,'C'),       (0x05,'3'),       False, 1),
	ce('Capricoin',             'CPC',     0x9c,   (0x1c,'C'),       None,             False, 0),
	ce('CashCoin',              'CASH',    0xa2,   (0x22,('E','F')), None,             False, 0),
	ce('CashOut',               'CSH',     0xa2,   (0x22,('E','F')), None,             False, 0),
	ce('ChainCoin',             'CHC',     0x9c,   (0x1c,'C'),       None,             False, 0),
	ce('Clams',                 'CLAM',    0x85,   (0x89,'x'),       (0x0d,'6'),       False, 1),
	ce('CoinMagi',              'XMG',     0x94,   (0x14,'9'),       None,             False, 0),
	ce('Condensate',            'RAIN',    0xbc,   (0x3c,'R'),       None,             False, 0),
	ce('CryptoBullion',         'CBX',     0x8b,   (0x0b,'5'),       None,             False, 0),
	ce('Cryptonite',            'XCN',     0x80,   (0x1c,'C'),       None,             False, 0),
	ce('CryptoPennies',         'CRPS',    0xc2,   (0x42,'T'),       None,             False, 0),
	ce('Dash',                  'DASH',    0xcc,   (0x4c,'X'),       (0x10,'7'),       False, 2),
	ce('Decred',                'DCR',     0x22de, (0x073f,'D'),     (0x071a,'D'),     False, 1),
	ce('DeepOnion',             'ONION',   0x9f,   (0x1f,'D'),       None,             False, 1),
	ce('Defcoin',               'DFC',     0x9e,   (0x1e,'D'),       (0x05,'3'),       False, 1),
	ce('Devcoin',               'DVC',     0x80,   (0x00,'1'),       None,             False, 1),
	ce('DigiByte',              'DGB',     0x80,   (0x1e,'D'),       (0x05,'3'),       False, 1),
	ce('DigiCoin',              'DGC',     0x9e,   (0x1e,'D'),       (0x05,'3'),       False, 1),
	ce('DogecoinDark',          'DOGED',   0x9e,   (0x1e,'D'),       (0x21,'E'),       False, 1),
	ce('Dogecoin',              'DOGE',    0x9e,   (0x1e,'D'),       (0x16,('9','A')), False, 2),
	ce('DopeCoin',              'DOPE',    0x88,   (0x08,'4'),       (0x05,'3'),       False, 1),
	ce('EGulden',               'EFL',     0xb0,   (0x30,'L'),       (0x05,'3'),       False, 1),
	ce('Emerald',               'EMD',     0xa2,   (0x22,('E','F')), None,             False, 0),
	ce('Emercoin',              'EMC',     0x80,   (0x21,'E'),       (0x5c,'e'),       False, 2),
	ce('EnergyCoin',            'ENRG',    0xdc,   (0x5c,'e'),       None,             False, 0),
	ce('Espers',                'ESP',     0xa1,   (0x21,'E'),       None,             False, 0),
	ce('Faircoin',              'FAI',     0xdf,   (0x5f,'f'),       (0x24,'F'),       False, 1),
	ce('Fastcoin',              'FST',     0xe0,   (0x60,('f','g')), None,             False, 0),
	ce('Feathercoin',           'FTC',     0x8e,   (0x0e,('6','7')), (0x05,'3'),       False, 2),
	ce('Fibre',                 'FIBRE',   0xa3,   (0x23,'F'),       None,             False, 0),
	ce('FlorinCoin',            'FLO',     0xb0,   (0x23,'F'),       None,             False, 0),
	ce('Fluttercoin',           'FLT',     0xa3,   (0x23,'F'),       None,             False, 0),
	ce('Fuel2Coin',             'FC2',     0x80,   (0x24,'F'),       None,             False, 0),
	ce('Fujicoin',              'FJC',     0xa4,   (0x24,'F'),       None,             False, 0),
	ce('Fujinto',               'NTO',     0xa4,   (0x24,'F'),       None,             False, 0),
	ce('GlobalBoost',           'BSTY',    0xa6,   (0x26,'G'),       None,             False, 0),
	ce('GlobalCurrencyReserve', 'GCR',     0x9a,   (0x26,'G'),       (0x61,'g'),       False, 1),
	ce('GoldenBird',            'XGB',     0xaf,   (0x2f,('K','L')), None,             False, 0),
	ce('Goodcoin',              'GOOD',    0xa6,   (0x26,'G'),       None,             False, 0),
	ce('GridcoinResearch',      'GRC',     0xbe,   (0x3e,('R','S')), None,             False, 1),
	ce('Gulden',                'NLG',     0xa6,   (0x26,'G'),       None,             False, 1),
	ce('Guncoin',               'GUN',     0xa7,   (0x27,('G','H')), None,             False, 1),
	ce('HamRadioCoin',          'HAM',     0x80,   (0x00,'1'),       None,             False, 1),
	ce('HTML5Coin',             'HTML5',   0xa8,   (0x28,'H'),       None,             False, 0),
	ce('HyperStake',            'HYP',     0xf5,   (0x75,'p'),       None,             False, 0),
	ce('iCash',                 'ICASH',   0xcc,   (0x66,'i'),       None,             False, 0),
	ce('ImperiumCoin',          'IPC',     0xb0,   (0x30,'L'),       None,             False, 0),
	ce('IncaKoin',              'NKA',     0xb5,   (0x35,'N'),       None,             False, 0),
	ce('Influxcoin',            'INFX',    0xe6,   (0x66,'i'),       None,             False, 0),
	ce('InPay',                 'INPAY',   0xb7,   (0x37,'P'),       None,             False, 0),
#	ce('iXcoin',                'IXC',     0x80,   (0x8a,'x'),       None,             False, 1),
	ce('Judgecoin',             'JUDGE',   0xab,   (0x2b,'J'),       None,             False, 0),
	ce('Jumbucks',              'JBS',     0xab,   (0x2b,'J'),       (0x69,'j'),       False, 2),
	ce('Lanacoin',              'LANA',    0xb0,   (0x30,'L'),       None,             False, 0),
	ce('Latium',                'LAT',     0x80,   (0x17,'A'),       None,             False, 0),
	ce('Litecoin',              'LTC',     0xb0,   (0x30,'L'),       (0x32,'M'),       True,  5), # old p2sh: 0x05
	ce('LiteDoge',              'LDOGE',   0xab,   (0x5a,'d'),       None,             False, 0),
	ce('LomoCoin',              'LMC',     0xb0,   (0x30,'L'),       None,             False, 0),
	ce('Marscoin',              'MARS',    0xb2,   (0x32,'M'),       None,             False, 0),
	ce('MarsCoin',              'MRS',     0xb2,   (0x32,'M'),       None,             False, 0),
	ce('MartexCoin',            'MXT',     0xb2,   (0x32,'M'),       None,             False, 0),
	ce('MasterCar',             'MCAR',    0xe6,   (0x17,'A'),       None,             False, 0),
	ce('MazaCoin',              'MZC',     0xe0,   (0x32,'M'),       (0x09,('4','5')), False, 2),
	ce('MegaCoin',              'MEC',     0xb2,   (0x32,'M'),       None,             False, 1),
	ce('MintCoin',              'MINT',    0xb3,   (0x33,'M'),       None,             False, 0),
	ce('Mobius',                'MOBI',    0x80,   (0x00,'1'),       None,             False, 0),
	ce('MonaCoin',              'MONA',    0xb0,   (0x32,'M'),       (0x05,'3'),       False, 1),
	ce('MonetaryUnit',          'MUE',     0x8f,   (0x0f,'7'),       (0x09,('4','5')), False, 1),
	ce('MoonCoin',              'MOON',    0x83,   (0x03,'2'),       None,             False, 0),
	ce('MyriadCoin',            'MYR',     0xb2,   (0x32,'M'),       (0x09,('4','5')), False, 1),
	ce('Myriadcoin',            'MYRIAD',  0xb2,   (0x32,'M'),       None,             False, 1),
	ce('Namecoin',              'NMC',     0xb4,   (0x34,('M','N')), (0x0d,'6'),       False, 1),
	ce('Neoscoin',              'NEOS',    0xef,   (0x3f,'S'),       (0xbc,'2'),       False, 1),
	ce('NevaCoin',              'NEVA',    0xb1,   (0x35,'N'),       None,             False, 0),
	ce('Novacoin',              'NVC',     0x88,   (0x08,'4'),       (0x14,'9'),       False, 1),
	ce('OKCash',                'OK',      0xb7,   (0x37,'P'),       (0x1c,'C'),       False, 1),
	ce('Omnicoin',              'OMC',     0xf3,   (0x73,'o'),       None,             False, 1),
	ce('Omni',                  'OMNI',    0xf3,   (0x73,'o'),       None,             False, 0),
	ce('Onix',                  'ONX',     0x80,   (0x8a,'x'),       None,             False, 0),
	ce('PandaCoin',             'PND',     0xb7,   (0x37,'P'),       (0x16,('9','A')), False, 1),
	ce('ParkByte',              'PKB',     0xb7,   (0x37,'P'),       (0x1c,'C'),       False, 1),
	ce('Particl',               'PART',    0x6c,   (0x38,'P'),       None,             False, 0),
	ce('Paycoin',               'CON',     0xb7,   (0x37,'P'),       None,             False, 1),
	ce('Peercoin',              'PPC',     0xb7,   (0x37,'P'),       (0x75,'p'),       False, 1),
	ce('PesetaCoin',            'PTC',     0xaf,   (0x2f,('K','L')), None,             False, 1),
	ce('PhoenixCoin',           'PXC',     0xb8,   (0x38,'P'),       None,             False, 0),
	ce('PinkCoin',              'PINK',    0x83,   (0x03,'2'),       None,             False, 1),
	ce('PIVX',                  'PIVX',    0xd4,   (0x1e,'D'),       None,             False, 0),
	ce('PokeChain',             'XPOKE',   0x9c,   (0x1c,'C'),       None,             False, 0),
	ce('Potcoin',               'POT',     0xb7,   (0x37,'P'),       (0x05,'3'),       False, 1),
	ce('Primecoin',             'XPM',     0x97,   (0x17,'A'),       (0x53,'a'),       False, 1),
	ce('Quark',                 'QRK',     0xba,   (0x3a,'Q'),       None,             False, 0),
	ce('ReddCoin',              'RDD',     0xbd,   (0x3d,'R'),       None,             False, 1),
	ce('Riecoin',               'RIC',     0x80,   (0x3c,'R'),       (0x05,'3'),       False, 2),
	ce('Rimbit',                'RBT',     0xbc,   (0x3c,'R'),       None,             False, 0),
	ce('Rubycoin',              'RBY',     0xbd,   (0x3d,'R'),       (0x55,'b'),       False, 1),
	ce('ShadowCash',            'SDC',     0xbf,   (0x3f,'S'),       (0x7d,'s'),       False, 1),
	ce('Sibcoin',               'SIB',     0x80,   (0x3f,'S'),       None,             False, 0),
	ce('SixEleven',             '611',     0x80,   (0x34,('M','N')), None,             False, 0),
	ce('SmileyCoin',            'SMLY',    0x99,   (0x19,'B'),       None,             False, 0),
	ce('Songcoin',              'SONG',    0xbf,   (0x3f,'S'),       None,             False, 0),
	ce('Spreadcoin',            'SPR',     0xbf,   (0x3f,'S'),       None,             False, 1),
	ce('Startcoin',             'START',   0xfd,   (0x7d,'s'),       (0x05,'3'),       False, 1),
	ce('StealthCoin',           'XST',     0xbe,   (0x3e,('R','S')), None,             False, 0),
	ce('SwagBucks',             'BUCKS',   0x99,   (0x3f,'S'),       None,             False, 0),
	ce('SysCoin',               'SYS',     0x80,   (0x00,'1'),       None,             False, 0),
	ce('TajCoin',               'TAJ',     0x6f,   (0x41,'T'),       None,             False, 0),
	ce('Templecoin',            'TPC',     0xc1,   (0x41,'T'),       (0x05,'3'),       False, 1),
	ce('Terracoin',             'TRC',     0x80,   (0x00,'1'),       None,             False, 0),
	ce('Titcoin',               'TIT',     0x80,   (0x00,'1'),       None,             False, 0),
	ce('TittieCoin',            'TTC',     0xc1,   (0x41,'T'),       None,             False, 0),
	ce('Transfer',              'TX',      0x99,   (0x42,'T'),       None,             False, 0),
	ce('Unobtanium',            'UNO',     0xe0,   (0x82,'u'),       (0x1e,'D'),       False, 2),
	ce('Vcash',                 'XVC',     0xc7,   (0x47,'V'),       None,             False, 0),
	ce('Vertcoin',              'VTC',     0xc7,   (0x47,'V'),       (0x05,'3'),       False, 1),
	ce('Viacoin',               'VIA',     0xc7,   (0x47,'V'),       (0x21,'E'),       False, 2),
	ce('VpnCoin',               'VPN',     0xc7,   (0x47,'V'),       (0x05,'3'),       False, 1),
	ce('WankCoin',              'WKC',     0x80,   (0x00,'1'),       None,             False, 1),
	ce('WashingtonCoin',        'WASH',    0xc9,   (0x49,'W'),       None,             False, 0),
	ce('WeAreSatoshi',          'WSX',     0x97,   (0x87,'w'),       None,             False, 0),
	ce('WisdomCoin',            'WISC',    0x87,   (0x49,'W'),       None,             False, 0),
	ce('WorldCoin',             'WDC',     0xc9,   (0x49,'W'),       None,             False, 1),
	ce('XRealEstateDevcoin',    'XRED',    0x80,   (0x00,'1'),       None,             False, 0),
	ce('ZetaCoin',              'ZET',     0xe0,   (0x50,'Z'),       None,             False, 0),
	ce('ZiftrCoin',             'ZRC',     0xd0,   (0x50,'Z'),       (0x05,'3'),       False, 1),
	ce('ZLiteQubit',            'ZLQ',     0xe0,   (0x26,'G'),       None,             False, 0),
	ce('Zoomcoin',              'ZOOM',    0xe7,   (0x67,'i'),       (0x5c,'e'),       False, 1),
	)

	coin_constants['testnet'] = (
	ce('Bitcoin',     'BTC',   0xef,   (0x6f,('m','n')), (0xc4,'2'),       True,  5),
	ce('BitcoinCashNode','BCH',0xef,   (0x6f,('m','n')), (0xc4,'2'),       True,  5),
	ce('BitcoinGold', 'BCG',   0xef,   (0x6f,('m','n')), (0xc4,'2'),       True, -1),
	ce('Dash',        'DASH',  0xef,   (0x8c,'y'),       (0x13,('8','9')), False, 1),
	ce('Decred',      'DCR',   0x230e, (0x0f21,'T'),     (0x0e6c,'S'),     False, 1),
	ce('Dogecoin',    'DOGE',  0xf1,   (0x71,'n'),       (0xc4,'2'),       False, 2),
	ce('Feathercoin', 'FTC',   0xc1,   (0x41,'T'),       (0xc4,'2'),       False, 2),
	ce('Viacoin',     'VIA',   0xff,   (0x7f,'t'),       (0xc4,'2'),       False, 2),
	ce('Emercoin',    'EMC',   0xef,   (0x6f,('m','n')), (0xc4,'2'),       False, 2),
	ce('Litecoin',    'LTC',   0xef,   (0x6f,('m','n')), (0x3a,'Q'),       True,  5), # old p2sh: 0xc4
	)

	coin_sources = (
	('BTC',    'https://github.com/bitcoin/bitcoin/blob/master/src/chainparams.cpp'),
	('EMC',    'https://github.com/emercoin/emercoin/blob/master/src/chainparams.cpp'), # checked mn,tn
	('LTC',    'https://github.com/litecoin-project/litecoin/blob/master-0.10/src/chainparams.cpp'),
	('DOGE',   'https://github.com/dogecoin/dogecoin/blob/master/src/chainparams.cpp'),
	('RDD',    'https://github.com/reddcoin-project/reddcoin/blob/master/src/base58.h'),
	('DASH',   'https://github.com/dashpay/dash/blob/master/src/chainparams.cpp'),
	('PPC',    'https://github.com/belovachap/peercoin/blob/master/src/base58.h'),
	('NMC',    'https://github.com/domob1812/namecore/blob/master/src/chainparams.cpp'),
	('FTC',    'https://github.com/FeatherCoin/Feathercoin/blob/master-0.8/src/base58.h'),
	('BLK',    'https://github.com/rat4/blackcoin/blob/master/src/chainparams.cpp'),
	('NSR',    'https://nubits.com/nushares/introduction'),
	('NBT',    'https://bitbucket.org/JordanLeePeershares/nubit/NuBit / src /base58.h'),
	('MZC',    'https://github.com/MazaCoin/MazaCoin/blob/master/src/chainparams.cpp'),
	('VIA',    'https://github.com/viacoin/viacoin/blob/master/src/chainparams.cpp'),
	('RBY',    'https://github.com/rubycoinorg/rubycoin/blob/master/src/base58.h'),
	('GRS',    'https://github.com/GroestlCoin/groestlcoin/blob/master/src/groestlcoin.cpp'),
	('DGC',    'https://github.com/DGCDev/digitalcoin/blob/master/src/chainparams.cpp'),
	('CCN',    'https://github.com/Cannacoin-Project/Cannacoin/blob/Proof-of-Stake/src/base58.h'),
	('DGB',    'https://github.com/digibyte/digibyte/blob/master/src/chainparams.cpp'),
	('MONA',   'https://github.com/monacoinproject/monacoin/blob/master-0.10/src/chainparams.cpp'),
	('CLAM',   'https://github.com/nochowderforyou/clams/blob/master/src/chainparams.cpp'),
	('XPM',    'https://github.com/primecoin/primecoin/blob/master/src/base58.h'),
	('NEOS',   'https://github.com/bellacoin/neoscoin/blob/master/src/chainparams.cpp'),
	('JBS',    'https://github.com/jyap808/jumbucks/blob/master/src/base58.h'),
	('ZRC',    'https://github.com/ZiftrCOIN/ziftrcoin/blob/master/src/chainparams.cpp'),
	('VTC',    'https://github.com/vertcoin/vertcoin/blob/master/src/base58.h'),
	('NXT',    'https://bitbucket.org/JeanLucPicard/nxt/src and unofficial at https://github.com/Blackcomb/nxt'),
	('MUE',    'https://github.com/MonetaryUnit/MUE-Src/blob/master/src/chainparams.cpp'),
	('ZOOM',   'https://github.com/zoom-c/zoom/blob/master/src/base58.h'),
	('VPN',    'https://github.com/Bit-Net/VpnCoin/blob/master/src/base58.h'),
	('CDN',    'https://github.com/ThisIsOurCoin/canadaecoin/blob/master/src/base58.h'),
	('SDC',    'https://github.com/ShadowProject/shadow/blob/master/src/chainparams.cpp'),
	('PKB',    'https://github.com/parkbyte/ParkByte/blob/master/src/base58.h'),
	('PND',    'https://github.com/coinkeeper/2015-04-19_21-22_pandacoin/blob/master/src/base58.h'),
	('START',  'https://github.com/startcoin-project/startcoin/blob/master/src/base58.h'),
	('GCR',    'https://github.com/globalcurrencyreserve/gcr/blob/master/src/chainparams.cpp'),
	('NVC',    'https://github.com/novacoin-project/novacoin/blob/master/src/base58.h'),
	('AC',     'https://github.com/AsiaCoin/AsiaCoinFix/blob/master/src/base58.h'),
	('BTCD',   'https://github.com/jl777/btcd/blob/master/src/base58.h'),
	('DOPE',   'https://github.com/dopecoin-dev/DopeCoinV3/blob/master/src/base58.h'),
	('TPC',    'https://github.com/9cat/templecoin/blob/templecoin/src/base58.h'),
	('OK',     'https://github.com/okcashpro/okcash/blob/master/src/chainparams.cpp'),
	('DOGED',  'https://github.com/doged/dogedsource/blob/master/src/base58.h'),
	('EFL',    'https://github.com/Electronic-Gulden-Foundation/egulden/blob/master/src/base58.h'),
	('POT',    'https://github.com/potcoin/Potcoin/blob/master/src/base58.h'),
	)

	@classmethod
	def get_supported_coins(cls, network):
		return [e for e in cls.coin_constants[network] if e.trust_level != -1]

	@classmethod
	def get_entry(cls, coin, network):
		try:
			idx = [e.symbol for e in cls.coin_constants[network]].index(coin.upper())
		except:
			return None
		return cls.coin_constants[network][idx]

def make_proto(e, *, testnet=False):

	proto = ('X_' if e.name[0] in '0123456789' else '') + e.name + ('Testnet' if testnet else '')

	if hasattr(CoinProtocol, proto):
		return

	def num2hexstr(n):
		return '{:0{}x}'.format(n, (4, 2)[n < 256])

	setattr(
		CoinProtocol,
		proto,
		type(
			proto,
			(mainnet,),
			{
				'base_coin': e.symbol,
				'addr_ver_info': dict(
					[(num2hexstr(e.p2pkh_info[0]), 'p2pkh')] +
					([(num2hexstr(e.p2sh_info[0]), 'p2sh')] if e.p2sh_info else [])
				),
				'wif_ver_num': {'std': num2hexstr(e.wif_ver_num)},
				'mmtypes':    ('L', 'C', 'S') if e.has_segwit else ('L', 'C'),
				'dfl_mmtype': 'L',
				'mmcaps':     (),
			},
		)
	)

def init_genonly_altcoins(usr_coin=None, *, testnet=False):
	"""
	Initialize altcoin protocol class or classes for current network.
	If usr_coin is a core coin, initialization is skipped.
	If usr_coin has a trust level of -1, an exception is raised.
	If usr_coin is None, initializes all coins for current network with trust level >-1.
	Returns trust_level of usr_coin, or 0 (untrusted) if usr_coin is None.
	"""

	data = {'mainnet': (), 'testnet': ()}
	networks = ['mainnet'] + (['testnet'] if testnet else [])
	network = 'testnet' if testnet else 'mainnet'

	if usr_coin is None:
		for network in networks:
			data[network] = CoinInfo.get_supported_coins(network)
	else:
		if usr_coin.lower() in gc.core_coins: # core coin, so return immediately
			return CoinProtocol.coins[usr_coin.lower()].trust_level
		for network in networks:
			data[network] = (CoinInfo.get_entry(usr_coin, network),)

		cinfo = data[network][0]
		if not cinfo:
			raise ValueError(f'{usr_coin.upper()!r}: unrecognized coin for network {network.upper()}')
		if cinfo.trust_level == -1:
			raise ValueError(f'{usr_coin.upper()!r}: unsupported (disabled) coin for network {network.upper()}')

	for e in data['mainnet']:
		make_proto(e)

	for e in data['testnet']:
		make_proto(e, testnet=True)

	for e in data['mainnet']:
		if e.symbol.lower() in CoinProtocol.coins:
			continue
		CoinProtocol.coins[e.symbol.lower()] = CoinProtocol.proto_info(
			name        = 'X_'+e.name if e.name[0] in '0123456789' else e.name,
			trust_level = e.trust_level)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tw.unspent: Bitcoin base protocol tracking wallet unspent outputs class
"""

from ....obj import ImmutableAttr, ListItemAttr, HexStr
from ....tw.unspent import TwUnspentOutputs

from .view import BitcoinTwView

class BitcoinTwUnspentOutputs(BitcoinTwView, TwUnspentOutputs):

	class display_type(TwUnspentOutputs.display_type):

		class squeezed(TwUnspentOutputs.display_type.squeezed):
			cols = ('num', 'txid', 'vout', 'addr', 'mmid', 'comment', 'amt', 'amt2', 'date')

		class detail(TwUnspentOutputs.display_type.detail):
			cols = ('num', 'txid', 'vout', 'addr', 'mmid', 'amt', 'amt2', 'block', 'date_time', 'comment')

	class MMGenTwUnspentOutput(TwUnspentOutputs.MMGenTwUnspentOutput):
		# required by gen_unspent(); setting valid_attrs explicitly is also more efficient
		valid_attrs = {
			'txid',
			'vout',
			'amt',
			'amt2',
			'comment',
			'twmmid',
			'addr',
			'confs',
			'date',
			'scriptPubKey',
			'skip'}
		date         = ListItemAttr(int, typeconv=False, reassign_ok=True)
		scriptPubKey = ImmutableAttr(HexStr)

	has_age = True
	can_group = True
	disp_spc = 5
	vout_w = 4
	hdr_lbl = 'unspent outputs'
	desc = 'unspent outputs'
	item_desc = 'unspent output'
	item_desc_pl = 'unspent outputs'
	dump_fn_pfx = 'listunspent'
	prompt_fs_in = [
		'Sort options: [t]xid, [a]mount, [A]ge, a[d]dr, [M]mgen addr, [r]everse',
		'Column options: toggle [D]ays/date/confs/block, gr[o]up, show [m]mgen addr',
		'View options: pager [v]iew, [w]ide pager view{s}',
		'Actions: [q]uit menu, [p]rint, r[e]draw, add [l]abel:']
	prompt_fs_repl = {
		'BCH': (1, 'Column options: toggle [D]ate/confs, cas[h]addr, gr[o]up, show [m]mgen addr')
	}
	extra_key_mappings = {
		'D':'d_days',
		'o':'d_group',
		't':'s_txid',
		'A':'s_age'}

	sort_funcs = {
		'addr':   lambda i: '{} {:010} {:024.12f}'.format(i.addr, 0xffffffff - abs(i.confs), i.amt),
		'age':    lambda i: '{:010} {:024.12f}'.format(0xffffffff - abs(i.confs), i.amt),
		'amt':    lambda i: '{:024.12f} {:010} {}'.format(i.amt, 0xffffffff - abs(i.confs), i.addr),
		'txid':   lambda i: f'{i.txid} {i.vout:04}',
		'twmmid': lambda i: '{} {:010} {:024.12f}'.format(
			i.twmmid.sort_key, 0xffffffff - abs(i.confs), i.amt)}

	async def get_rpc_data(self):
		# bitcoin-cli help listunspent:
		# Arguments:
		# 1. minconf        (numeric, optional, default=1) The minimum confirmations to filter
		# 2. maxconf        (numeric, optional, default=9999999) The maximum confirmations to filter
		# 3. addresses      (json array, optional, default=empty array) A json array of bitcoin addresses
		# 4. include_unsafe (boolean, optional, default=true) Include outputs that are not safe to spend
		# 5. query_options  (json object, optional) JSON with query options

		# for now, self.addrs is just an empty list for Bitcoin and friends
		add_args = (9999999, self.addrs) if self.addrs else ()
		return await self.rpc.call('listunspent', self.minconf, *add_args)



#!/usr/bin/env python3

"""
Convert an MMGen 'v2' transaction file (amounts as BTCAmt()) to 'v3' (amounts as
strings).  Version 3 TX files were introduced with MMGen version 0.9.7
"""

import sys, os, asyncio

from mmgen.cfg import Config
from mmgen.tx import CompletedTX

repo_root = os.path.split(os.path.abspath(os.path.dirname(sys.argv[0])))[0]
sys.path = [repo_root] + sys.path

opts_data = {
	'text': {
		'desc':    "Convert an MMGen transaction file from v2 format to v3 format",
		'usage':   "<TX file>",
		'options': """
-h, --help     Print this help message
-d, --outdir=d Output files to directory 'd' instead of working dir
-q, --quiet    Write (and overwrite) files without prompting
-S, --stdout   Write data to STDOUT instead of file
"""
	}
}

cfg = Config(opts_data=opts_data)

if len(cfg._args) != 1:
	cfg._usage()

tx = asyncio.run(CompletedTX(cfg._args[0], quiet_open=True))
tx.file.write(ask_tty=False, ask_overwrite=not cfg.quiet, ask_write=not cfg.quiet)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
pyversion: Python version string operations
"""

class PythonVersion(str):

	major = 0
	minor = 0

	def __new__(cls, arg=None):
		if isinstance(arg, PythonVersion):
			return arg
		if arg:
			major, minor = arg.split('.')
		else:
			import platform
			major, minor = platform.python_version_tuple()[:2]
		me = str.__new__(cls, f'{major}.{minor}')
		me.major = int(major)
		me.minor = int(minor)
		return me

	def __lt__(self, other):
		other = type(self)(other)
		return self.major < other.major or (self.major == other.major and self.minor < other.minor)

	def __le__(self, other):
		other = type(self)(other)
		return self.major < other.major or (self.major == other.major and self.minor <= other.minor)

	def __eq__(self, other):
		other = type(self)(other)
		return self.major == other.major and self.minor == other.minor

	def __ne__(self, other):
		other = type(self)(other)
		return not (self.major == other.major and self.minor == other.minor)

	def __gt__(self, other):
		other = type(self)(other)
		return self.major > other.major or (self.major == other.major and self.minor > other.minor)

	def __ge__(self, other):
		other = type(self)(other)
		return self.major > other.major or (self.major == other.major and self.minor >= other.minor)

python_version = PythonVersion()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.tx.bump: Ethereum transaction bump class
"""

from decimal import Decimal

from ....tx import bump as TxBase
from .completed import Completed, TokenCompleted
from .new import TokenNew
from .new_swap import NewSwap
from .unsigned import AutomountUnsigned, TokenAutomountUnsigned

class Bump(Completed, NewSwap, TxBase.Bump):
	desc = 'fee-bumped transaction'

	def get_orig_rel_fee(self):
		return self.txobj['gasPrice'].to_unit('Gwei')

	@property
	def min_fee(self):
		return self.fee * Decimal('1.101')

	def bump_fee(self, idx, fee):
		self.txobj['gasPrice'] = self.fee_abs2gasprice(fee)

	async def get_nonce(self):
		return self.txobj['nonce']

class TokenBump(TokenCompleted, TokenNew, Bump):
	desc = 'fee-bumped transaction'

class AutomountBump(Bump):
	ext       = AutomountUnsigned.ext
	automount = AutomountUnsigned.automount

class TokenAutomountBump(TokenBump):
	ext       = TokenAutomountUnsigned.ext
	automount = TokenAutomountUnsigned.automount
